import logging
import sys

from ..toolkit.app import SubCommand
from .app import app

logger = logging.getLogger(__name__)


@app.command
class Discover(SubCommand):
    """Introspect system and PostgreSQL instance."""

    def main(self, args):
        self.app.discover.refresh()
        self.app.discover.write()
        self.app.discover.write(sys.stdout)
        logger.debug("Discover etag is %s.", self.app.discover.etag)
        return 0



import datetime
import logging
import sys
import time
from logging.config import dictConfig
from logging.handlers import SysLogHandler


class ColoredStreamHandler(logging.StreamHandler):
    _color_map = {  # Colors code from systemctl/journalctl
        logging.DEBUG: "2",
        logging.INFO: "1",
        logging.WARN: "38;5;185",
        logging.ERROR: "31",
        logging.CRITICAL: "1;91",
    }

    def format(self, record):
        lines = logging.StreamHandler.format(self, record)
        color = self._color_map.get(record.levelno, "39")
        lines = "".join(
            [f"\033[0;{color}m{line}\033[0m" for line in lines.splitlines(True)]
        )
        return lines


class MultilineFormatter(logging.Formatter):
    def format(self, record):
        s = logging.Formatter.format(self, record)
        if "\n" in s:
            s = s.replace("\n", "\n\t")
        return s


class NullHandler(logging.Handler):
    def handle(self, record):
        pass

    def emit(self, record):
        pass

    def createLock(self):
        self.lock = None


class SystemdFormatter(MultilineFormatter):
    # cf. http://0pointer.de/blog/projects/journal-submit.html
    priority_map = {
        logging.NOTSET: 6,
        logging.DEBUG: 7,
        logging.INFO: 6,
        logging.WARNING: 4,
        logging.ERROR: 3,
        logging.CRITICAL: 2,
    }

    def format(self, record):
        s = MultilineFormatter.format(self, record)
        prefix = "<%d>" % self.priority_map[record.levelno]
        lines = [prefix + line for line in s.splitlines(True)]
        return "".join(lines)


class LastnameFilter(logging.Filter):
    root, _, _ = __name__.partition(".")

    def filter(self, record):
        record.lastname = record.name
        if record.name.startswith(self.root + "."):
            _, record.lastname = record.name.rsplit(".", 1)
        # Always log, we are just enriching records.
        return 1


HANDLERS = {
    "file": {"()": "logging.FileHandler", "mode": "a", "formatter": "dated_syslog"},
    "syslog": {"()": "logging.handlers.SysLogHandler", "formatter": "syslog"},
    "stderr": {"()": __name__ + ".NullHandler"},
}


def setup_logging(**kw):
    logging_config = generate_logging_config(**kw)
    dictConfig(logging_config)


def configure_debug(logging_config, core, debug):
    # If --debug or DEBUG=1, apply DEBUG to all core loggers
    if debug in (True, "__debug__"):
        debug = core

    if hasattr(debug, "split"):
        debug = filter(None, debug.split(","))

    # Now apply debug level.
    if debug:
        for loggername in debug:
            logger = logging_config["loggers"].setdefault(loggername, {})
            logger["level"] = "DEBUG"


def generate_logging_config(
    level=None,
    destination=None,
    facility="local0",
    method="stderr",
    debug=None,
    systemd=False,
    **kw,
):
    # Our logging strategy is to always log on stderr and allow to log on
    # either syslog or a file, depending on user configuration.
    #
    # stderr mean either console when launching in a terminal, file when stderr
    # is piped or systemd output. Thus we adapt stderr to these cases depending
    # on systemd an isatty().

    core = LastnameFilter.root

    if level is None:
        level = "DEBUG" if debug else "INFO"

    if debug is None:
        debug = level == "DEBUG"

    verbose = debug or level == "DEBUG"

    facility = SysLogHandler.facility_names[facility]
    HANDLERS["syslog"]["facility"] = facility
    HANDLERS["syslog"]["address"] = destination
    HANDLERS["file"]["filename"] = destination

    stderr_handler = "logging.StreamHandler"
    if sys.stderr.isatty():
        stderr_handler = __name__ + ".ColoredStreamHandler"
        datefmt = "%H:%M:%S"
    else:
        datefmt = None

    minimal_fmt = "%(levelname)s:  %(lastname)s: %(message)s"
    verbose_fmt = "%(asctime)s " + core + "[%(process)d] " + minimal_fmt
    syslog_fmt = core + "[%(process)d] %(levelname)s:  %(lastname)s: %(message)s"

    logging_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "filters": {"lastname": {"()": __name__ + ".LastnameFilter"}},
        "formatters": {
            "console": {
                "()": __name__ + ".MultilineFormatter",
                "format": verbose_fmt if verbose else minimal_fmt,
                "datefmt": datefmt,
            },
            "dated_syslog": {
                "()": __name__ + ".MultilineFormatter",
                "format": "%(asctime)s " + syslog_fmt,
            },
            "syslog": {"()": __name__ + ".MultilineFormatter", "format": syslog_fmt},
            "systemd": {
                "()": __name__ + ".SystemdFormatter",
                "format": "%(lastname)s: %(message)s",
            },
        },
        "handlers": {
            # Avoid instanciate all handlers, especially syslog which opens /dev/log.
            "configured": dict(HANDLERS[method], filters=["lastname"]),
            "stderr": {
                "()": stderr_handler,
                "formatter": "systemd" if systemd else "console",
                "filters": ["lastname"],
            },
        },
        "root": {"level": "INFO", "handlers": ["stderr", "configured"]},
        "loggers": {},
    }

    # Apply level to temboard loggers only
    logging_config["loggers"][core] = dict(level=level)

    configure_debug(logging_config, core, debug)

    return logging_config


def localoffset():
    now_utc = datetime.datetime.now(datetime.timezone.utc)
    now_local = now_utc.astimezone()
    offset = now_local - now_utc
    hours = offset.seconds // 3600
    minutes = (offset.seconds // 60) % 60
    return "+%02d%02d" % (hours, minutes)


# Make logging format consistent with Prometheus ts= format.
logging.Formatter.converter = time.gmtime
logging.Formatter.default_time_format = "%Y-%m-%dT%H:%M:%S"
logging.Formatter.default_msec_format = "%s.%03dZ"  # Append UTC timezone.



from textwrap import dedent


def get_settings_categories(conn):
    return {
        "categories": [
            row["category"]
            for row in conn.query("""\
    SELECT DISTINCT(category) FROM pg_settings ORDER BY category
    """)
        ]
    }


def get_setting(conn, name):
    return conn.queryscalar("SELECT setting FROM pg_settings WHERE name = %s", (name,))


def get_settings(conn, category=None, search=None):
    where = ""
    if search:
        where = dedent("""\
        WHERE name ILIKE '%{0}%'
         OR short_desc ILIKE '%{0}%'
         OR extra_desc ILIKE '%{0}%'
        """).format(search)
    query = (
        dedent("""\
    SELECT
      name,
      setting,
      current_setting(name) AS current_setting,
      unit,
      vartype,
      min_val, max_val, enumvals,
      context, category,
      short_desc || ' ' || coalesce(extra_desc, '') AS desc,
      boot_val, reset_val,
      pending_restart
    FROM pg_settings
    %s
    ORDER BY category, name
    """)
        % where
    )

    ret = {}
    for row in conn.query(query):
        if category and category != row["category"]:
            continue

        rows = ret.setdefault(row["category"], [])
        enumvals = row["enumvals"]
        if enumvals is not None:
            # format enumvals as before switching from tpc to psycopg2
            enumvals = "{%s}" % ",".join(enumvals)
        rows.append(
            {
                "name": row["name"],
                "setting": row["setting"],
                "setting_raw": row["current_setting"],
                "unit": row["unit"],
                "vartype": row["vartype"],
                "min_val": row["min_val"],
                "max_val": row["max_val"],
                "boot_val": row["boot_val"],
                "reset_val": row["reset_val"],
                "enumvals": enumvals,
                "context": row["context"],
                "desc": row["desc"],
                "pending_restart": row["pending_restart"],
            }
        )

    return [{"category": k, "rows": v} for k, v in ret.items()]


def get_settings_status(conn):
    settings = get_settings(conn)
    pending_restart_changes = []
    pending_restart = False
    for category in settings:
        for row in category["rows"]:
            if row["pending_restart"]:
                pending_restart = True
                pending_restart_changes.append(row)
    return {
        "restart_pending": pending_restart,
        "restart_changes": pending_restart_changes,
    }



import logging
import os.path
from urllib.parse import urlparse

from ..toolkit.app import SubCommand
from ..toolkit.errors import UserError
from ..toolkit.http import TemboardClient
from ..toolkit.signing import load_public_key
from .app import app

logger = logging.getLogger(__name__)


@app.command
class FetchKey(SubCommand):
    """Fetch signing public key."""

    name = "fetch-key"

    def define_arguments(self, parser):
        parser.add_argument(
            "--force",
            action="store_true",
            default=False,
            help="Force overwriting existing files.",
        )

    def main(self, args):
        pub = self.app.config.temboard.signing_public_key
        if os.path.exists(pub) and not args.force:
            raise UserError("%s exists. Use --force to overwrite." % pub)

        ui_url_raw = self.app.config.temboard.ui_url.rstrip("/")
        ui_url = urlparse(ui_url_raw)
        ui_client = TemboardClient.factory(
            self.app.config,
            scheme=ui_url.scheme,
            host=ui_url.hostname,
            port=ui_url.port,
        )

        logger.info("Requesting public key from %s.", ui_url_raw)
        response = ui_client.get("/signing.key")
        response.raise_for_status()
        pem = response.read()

        logger.info("Validating PEM data.")
        load_public_key(pem)

        with open(pub, "wb") as fo:
            fo.write(pem)

        logger.info("%s updated from %s.", pub, ui_url_raw)
        logger.info("Please reload agent service if running.")

        return 0



import logging
import ssl
from socket import error as SocketError
from wsgiref.simple_server import ServerHandler, WSGIRequestHandler, make_server

from bottle import debug, default_app

from .. import __version__
from ..errors import UserError
from ..toolkit import syncio

logger = logging.getLogger(__package__)


class HTTPDService(syncio.Service):
    name = "web"

    def __init__(self, app):
        self.app = app

    def __str__(self):
        return self.name

    # for services.run
    def setup(self, *_, **__):
        ServerHandler.server_software = "temBoard-agent/%s" % __version__

        try:
            bottle = default_app()
            debug(self.app.debug)

            self.server = make_server(
                self.app.config.temboard.address,
                self.app.config.temboard.port,
                app=bottle,
                handler_class=CustomWSGIRequestHandler,
            )
        except SocketError as e:
            raise UserError(f"Failed to start HTTPS server: {e}.")
        try:
            logger.debug("Using SSL key %s.", self.app.config.temboard.ssl_key_file)
            logger.debug(
                "Using SSL certificate %s.", self.app.config.temboard.ssl_cert_file
            )
            ctx = ssl.SSLContext()
            ctx.load_cert_chain(
                self.app.config.temboard.ssl_cert_file,
                self.app.config.temboard.ssl_key_file,
            )
            ctx.set_ciphers(
                ":".join(
                    [
                        # From Mozilla SSL configuration generator. 2023-07-28
                        "ECDHE-ECDSA-AES128-GCM-SHA256",
                        "ECDHE-RSA-AES128-GCM-SHA256",
                        "ECDHE-ECDSA-AES256-GCM-SHA384",
                        "ECDHE-RSA-AES256-GCM-SHA384",
                        "ECDHE-ECDSA-CHACHA20-POLY1305",
                        "ECDHE-RSA-CHACHA20-POLY1305",
                        "DHE-RSA-AES128-GCM-SHA256",
                        "DHE-RSA-AES256-GCM-SHA384",
                        "DHE-RSA-CHACHA20-POLY1305",
                    ]
                )
            )

            self.server.socket = ctx.wrap_socket(self.server.socket, server_side=True)
        except Exception as e:
            raise UserError(f"Failed to setup SSL: {e}.")
        self.server.timeout = 1

        logger.info(
            "Serving at https://%s:%s.",
            self.app.config.temboard.address,
            self.app.config.temboard.port,
        )

    # for syncio.Loop
    def accept(self):
        self.server.handle_request()


class CustomWSGIRequestHandler(WSGIRequestHandler):
    def get_environ(self):
        env = super().get_environ()

        # Save raw PATH_INFO for signature computation.
        path, _, _ = self.path.partition("?")
        env["RAW_PATH_INFO"] = path

        return env

    def log_request(self, code="-", size="-"):
        if code > "50":
            logmethod = logger.error
        else:
            logmethod = logger.info
        logmethod(
            # Same format as Tornado access.
            "%s %s %s (%s) %s %s",
            code,
            self.command,  # Method
            self.path,
            self.address_string(),
            "-",  # Not timed.
            size,
        )



import sys

from bottle import default_app

from .cli.app import app
from .web.app import create_app


def main():
    default_app.pop()  # Remove default app.
    default_app.push(create_app(app))

    # Import core HTTP routes
    __import__(__package__ + ".web.core")

    # Import commands
    __import__(__package__ + ".cli.discover")
    __import__(__package__ + ".cli.fetch_key")
    __import__(__package__ + ".cli.register")
    __import__(__package__ + ".cli.routes")
    __import__(__package__ + ".cli.runscript")
    __import__(__package__ + ".cli.serve")
    __import__(__package__ + ".cli.tasks")
    __import__(__package__ + ".cli.web")
    return app()


if "__main__" == __name__:
    sys.exit(main())



import re

import psycopg2.sql
from bottle import HTTPError, default_app, request

from ...queries import QUERIES
from ...web.app import CustomBottle
from . import functions as pgconf_functions

bottle = CustomBottle()


@bottle.post("/reload")
def post_reload(pgconn):
    """Reload Postgres configuration."""
    default_app().push_audit_notification("PostgreSQL configuration reload.")
    pgconn.execute("SELECT pg_reload_conf();")


@bottle.get("/settings")
def get_settings(pgconn):
    """Return all settings metadata."""
    return list(pgconn.query(QUERIES["pgconf-settings"]))


@bottle.post("/settings")
def post_settings(pgconn, new=None):
    """Applies a JSON mapping of setting -> value."""
    new = request.json if new is None else new
    if not hasattr(new, "items"):
        raise HTTPError(406, "Requires a mapping of settings and values.")
    if not new:
        raise HTTPError(406, "No settings.")

    for name, setting in new.items():
        if setting is None:
            raise HTTPError(406, "Setting value is required.")
        if len(setting) > 1024:
            raise HTTPError(406, "Setting value is too long.")

        default_app().push_audit_notification(f"Setting {name} to {setting}.")
        sql = psycopg2.sql.SQL("""ALTER SYSTEM SET {} TO %(setting)s;""")
        try:
            pgconn.execute(
                sql.format(psycopg2.sql.Identifier(name)), {"setting": setting}
            )
        except psycopg2.DatabaseError as e:
            return HTTPError(406, e.pgerror)

    post_reload(pgconn)


@bottle.delete("/settings/<name>")
def delete_settings(pgconn, name):
    """Reset settings to default."""
    default_app().push_audit_notification(f"Reseting {name} to default.")
    sql = psycopg2.sql.SQL("""ALTER SYSTEM RESET {};""")
    try:
        pgconn.execute(sql.format(psycopg2.sql.Identifier(name)))
    except psycopg2.DatabaseError as e:
        return HTTPError(406, e.pgerror)

    post_reload(pgconn)


# DEPRECATED: Remove these routes once UI uses only the above.


@bottle.get("/configuration")
def get_configuration(pgconn):
    return get_configuration_category(pgconn, None)


@bottle.get("/configuration/category/<category:path>")
def get_configuration_category(pgconn, category):
    search = None
    if "filter" in request.query:
        if not re.match("([a-zA-Z0-9_]{3,128})", request.query["filter"]):
            raise HTTPError(406, "Parameter 'filter' is malformed")
        search = request.query["filter"]
    if category:
        # Unquote +.
        category = category.replace("+", " ")
    return pgconf_functions.get_settings(pgconn, category, search)


@bottle.get("/configuration/categories")
def get_configuration_categories(pgconn):
    return pgconf_functions.get_settings_categories(pgconn)


@bottle.post("/configuration")
def post_configuration(pgconn):
    if "settings" not in request.json:
        raise HTTPError(406, "Parameter 'settings' not sent.")
    reset = {i["name"] for i in request.json["settings"] if not i["setting"]}
    for name in reset:
        out = delete_settings(pgconn, name)
        if out:
            return out

    # Transpose to new payload format.
    new = {
        i["name"]: i["setting"]
        for i in request.json["settings"]
        if i["name"] not in reset
    }
    if new:
        out = post_settings(pgconn, new)

    if out is None:
        out = {"settings": []}
    return out


@bottle.get("/configuration/status")
def get_status(pgconn):
    return pgconf_functions.get_settings_status(pgconn)



# User Input Validator
#
# These functions provide stateless validation of user input, mainly CLI
# arguments and environment variables.
#
# On invalid input, a ValueError is raised. Other exceptions are considered a
# bug.
#
# A validator is idempotent. It must accepts what it returns.

import json
import logging
import os.path
import re
from logging.handlers import SysLogHandler
from urllib.parse import urlparse

from .log import HANDLERS as LOG_METHODS
from .utils import strtobool

_address_re = re.compile(
    r"(?:[3-9]\d?|2(?:5[0-5]|[0-4]?\d)?|1\d{0,2}|\d)"
    r"(\.(?:[3-9]\d?|2(?:5[0-5]|[0-4]?\d)?|1\d{0,2}|\d"
    r")){3}$"
)


def address(raw):
    if not _address_re.match(raw):
        raise ValueError("invalid address")
    return raw


def boolean(raw):
    if raw in (True, False):
        return raw

    return strtobool(raw)


def dir_(raw):
    raw = os.path.realpath(raw)
    if not os.path.isdir(raw):
        raise ValueError("Not a directory")
    return raw


_email_re = re.compile(r"^([a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+)$")


def email(raw):
    if not _email_re.match(raw):
        raise ValueError("invalid email")
    return raw


def file_(raw):
    if not raw:
        return raw
    raw = os.path.realpath(raw)
    if not os.path.exists(raw):
        raise ValueError("%s: File not found" % raw)
    return raw


def password(raw):
    if len(raw) < 8:
        raise ValueError("Password must be at least 8 characters long")
    return raw


def path(raw):
    if not raw:
        return raw
    raw = os.path.realpath(raw)
    parent = os.path.dirname(raw)
    if not os.path.isdir(parent):
        raise ValueError("Missing parent directory of: %s" % raw)
    return raw


def phone(raw):
    if not re.match(r"\+[0-9]+", raw):
        raise ValueError("invalid phone number")
    return raw


def fqdn(raw):
    if "\n" in raw:
        raise ValueError("New line in FQDN.")
    if not re.match(
        r"(?=^.{1,253}$)"  # check length between 4 and 253
        r"(^((?!-)[a-zA-Z0-9-]{1,63}(?<!-))"
        r"(\.(?!-)[a-zA-Z0-9-]{1,63}(?<!-))*$)",
        raw,
    ):
        raise ValueError("%s is not an FQDN" % raw)
    return raw


_identifier_re = re.compile(r"^[a-zA-Z0-9]+$")


def jsonlist(raw):
    if hasattr(raw, "lower"):
        raw = json.loads(raw)

    if not isinstance(raw, list):
        raise ValueError("not a list")

    raw = [str(e) for e in raw]
    for entry in raw:
        if not _identifier_re.match(entry):
            raise ValueError("%s is invalid" % entry)

    return raw


def port(raw):
    port = int(raw)

    if 0 > port or port > 65635:
        raise ValueError("Port out of range")

    return port


def loglevel(raw):
    raw = raw.upper()
    levelnames = logging._nameToLevel
    if raw not in levelnames:
        raise ValueError("unkown log level")
    return raw


def logmethod(raw):
    if raw not in LOG_METHODS:
        raise ValueError(f"unknown logging method {raw}")
    return raw


def syslogfacility(raw):
    if raw not in SysLogHandler.facility_names:
        raise ValueError("unkown syslog facility")
    return raw


def writeabledir(raw):
    raw = dir_(raw)
    if not os.access(raw, os.W_OK):
        raise ValueError("Not writable")
    return raw


def commalist(raw):
    return list(filter(None, [w.strip() for w in raw.split(",")]))


def nday(raw):
    nday = int(raw)

    if nday < 1:
        raise ValueError("Number of day not valid")

    return nday


def url(raw):
    url = urlparse(raw)
    if not url.scheme.startswith("http"):
        raise ValueError("HTTP URL required")
    if not url.netloc:
        raise ValueError("Missing host and port")
    return raw


def slug(raw):
    if not re.match(r"^[a-z0-9_\-.]{3,24}$", raw):
        raise ValueError(
            "identifier must be 3-24 characters long, lowercase or numbers, and may contain point, hyphen or underscore"
        )
    return raw



def remove_passwords(instances):
    clean_instances = []
    for instance in instances:
        clean_instance = {}
        for k in instance.keys():
            if k != "password":
                clean_instance[k] = instance[k]
        clean_instances.append(clean_instance)
    return clean_instances



import time

# Label returned when the data is not available
NotAvailableLabel = "N/A"


def bytes2human(num):
    """
    Convert a size into a human readable format.
    """
    symbols = ("K", "M", "G", "T", "P", "E", "Z", "Y")
    prefix = {}
    nume = ""
    if num < 0:
        num = num * -1
        nume = "-"
    for pos, sym in enumerate(symbols):
        prefix[sym] = 1 << (pos + 1) * 10
    for sym in reversed(symbols):
        if num >= prefix[sym]:
            value = "%.2f" % float(float(num) / float(prefix[sym]))
            return f"{nume}{value}{sym}"
    return f"{nume}{num:.2f}B"


def memory_total_size():
    """
    Get the total amount of memory from /proc/meminfo.
    """
    try:
        with open("/proc/meminfo") as fd:
            for line in fd.readlines():
                items = line.split()
                if len(items) > 0:
                    if items[0][:-1] == "MemTotal":
                        mem_total = int(items[1])
                        if items[2] == "kB":
                            return mem_total * 1024
                        else:
                            return mem_total
            fd.close()
    except Exception:
        pass
    return


class Process:
    """
    Process class aimed to retreive process informations such CPU usage, memory
    usage, waiting in uninterruptible disk sleep (IO wait), disk read and write
    rates.
    """

    def __init__(self, pid, mem_total, page_size):
        """
        Constructor.
        """
        self.pid = pid
        self.iow = False
        self.rss = 0
        self.cpu_time = 0
        self.cpu_time_capture = 0
        self.page_size = page_size
        self.mem_total = mem_total
        self.read_bytes = 0
        self.write_bytes = 0
        self.io_capture = 0
        if self.pid:
            self.__init_stat()
            self.__parse_proc_io()

    def __parse_proc_stat(self):
        """
        /proc/<pid>/stat parser for getting CPU times, RSS and IOW.
        """
        iow = "N"
        cpu_time = 0
        cpu_time_capture = 0
        rss = 0
        try:
            with open("/proc/%s/stat" % (self.pid)) as fd:
                lines = fd.read()
                infos = lines[:-1].split()
                # IO wait
                if infos[2] == "D":
                    iow = "Y"
                # RSS
                rss = infos[23]
                # CPU time
                cpu_time = (
                    float(infos[13])
                    + float(infos[14])
                    + float(infos[15])
                    + float(infos[16])
                )
                cpu_time_capture = time.time()
            return (iow, rss, cpu_time, cpu_time_capture)
        except Exception:
            # Case when /proc/<pid>/stat can be read/parsed
            return (NotAvailableLabel, None, None, None)

    def __parse_proc_io(self):
        """
        /proc/<pid>/io parser for getting read_bytes and write_bytes.
        """
        read_bytes = 0
        write_bytes = 0
        io_capture = 0
        try:
            with open("/proc/%s/io" % (self.pid)) as fd:
                io_capture = time.time()
                for line in fd.readlines():
                    infos = line.split()
                    if infos[0][:-1] == "read_bytes":
                        read_bytes = int(infos[1])
                    if infos[0][:-1] == "write_bytes":
                        write_bytes = int(infos[1])
            return (read_bytes, write_bytes, io_capture)
        except Exception:
            # Case when /proc/<pid>/io can be read/parsed
            return (None, None, None)

    def __init_stat(self):
        (self.iow, self.rss, self.cpu_time, self.cpu_time_capture) = (
            self.__parse_proc_stat()
        )

        (self.read_bytes, self.write_bytes, self.io_capture) = self.__parse_proc_io()

    def cpu_usage(self):
        """
        Calculates and returns CPU usage based on CPU time
        """
        (_, _, new_cpu_time, new_cpu_time_capture) = self.__parse_proc_stat()
        if new_cpu_time is not None and new_cpu_time_capture is not None:
            cpu_usage = float(
                "%.2f"
                % (
                    round(
                        float(
                            float(new_cpu_time - self.cpu_time)
                            / float(new_cpu_time_capture - self.cpu_time_capture)
                        ),
                        2,
                    )
                )
            )
            self.cpu_time = new_cpu_time
            self.cpu_time_capture = new_cpu_time_capture
            return cpu_usage
        else:
            return NotAvailableLabel

    def mem_usage(self, new_capture=False):
        """
        Calculates and returns memory usage.
        """
        if new_capture is True:
            self.__init_stat()
        if self.rss is not None and self.mem_total is not None:
            return float(
                "%.2f"
                % (
                    round(
                        float(float(self.rss) * self.page_size)
                        / float(self.mem_total)
                        * 100,
                        2,
                    )
                )
            )
        else:
            return NotAvailableLabel

    def io_usage(self):
        """
        Calculates and returns I/O disk rates.
        """
        (new_read_bytes, new_write_bytes, new_io_capture) = self.__parse_proc_io()
        if (
            new_read_bytes is not None
            and new_write_bytes is not None
            and new_io_capture is not None
        ):
            read_rate = float(
                "%.2f"
                % (
                    round(
                        float(
                            float(new_read_bytes - self.read_bytes)
                            / float(new_io_capture - self.io_capture)
                        ),
                        2,
                    )
                )
            )
            write_rate = float(
                "%.2f"
                % (
                    round(
                        float(
                            float(new_write_bytes - self.write_bytes)
                            / float(new_io_capture - self.io_capture)
                        ),
                        2,
                    )
                )
            )
            self.read_bytes = new_read_bytes
            self.write_bytes = new_write_bytes
            self.io_capture = new_io_capture
            return (bytes2human(read_rate), bytes2human(write_rate))
        else:
            return (NotAvailableLabel, NotAvailableLabel)



# Note that our home-made background task implementation uses a different
# semantic that state of the art background task implemtation:
#
# a task function is called a worker
# a worker process is called workerpool
# a message is called a task
# the message broker is called taskmanager
#

import logging.config

from ..toolkit.app import SubCommand
from ..toolkit.errors import UserError
from ..toolkit.taskmanager import FlushTasksMixin, RunTaskMixin
from .app import app

logger = logging.getLogger(__name__)


@app.command
class Tasks(SubCommand):
    """Manage background tasks."""

    def main(self, args):
        raise UserError("Missing sub-command. See --help for details.")


@Tasks.command
class Flush(FlushTasksMixin, SubCommand):
    """Flush all tasks."""


@Tasks.command
class Run(RunTaskMixin, SubCommand):
    """Run a task foreground."""

    def main(self, args):
        workers = self.iter_workers()

        if "?" == args.worker_name:
            self.print_workers(workers)
        else:
            worker, worker_args = self.compute_worker_args(workers, args)
            worker(*worker_args)
        return 0



import logging
import time

from bottle import default_app

from ...toolkit import logfmt, taskmanager
from ...toolkit.configuration import OptionSpec
from ...toolkit.utils import utcnow
from ...web.app import CustomBottle
from . import db, metrics

bottle = CustomBottle()
logger = logging.getLogger(__name__)
workers = taskmanager.WorkerSet()


@bottle.get("/")
def dashboard(pgconn):
    out = metrics.get_metrics_queue(default_app().temboard.config)
    out["status"] = default_app().temboard.status.get(conn=pgconn)
    return out


@bottle.get("/config")
def dashboard_config():
    app = default_app().temboard
    return dict(
        scheduler_interval=app.config.dashboard.scheduler_interval,
        history_length=app.config.dashboard.history_length,
    )


@bottle.get("/history")
def dashboard_history():
    return metrics.get_history_metrics_queue(default_app().temboard.config)


@workers.register(pool_size=1)
def dashboard_collector_worker(app, pool=None):
    logger.info("Collecting metrics.")

    data = metrics.get_metrics(app, pool)

    # We don't want to store notifications in the history.
    data.pop("notifications", None)
    logger.debug(logfmt.format(**data))

    db.add_metric(
        app.config.temboard.home,
        "dashboard.db",
        time.time(),
        data,
        app.config.dashboard.history_length,
    )


BATCH_DURATION = 5 * 60  # 5 minutes


@workers.register(pool_size=1)
def dashboard_collector_batch_worker(app):
    # Loop each configured interval in the batch duration.
    interval = app.config.dashboard.scheduler_interval
    pool = None
    start = utcnow()
    elapsed = 0
    while elapsed < BATCH_DURATION:
        if elapsed > 0:
            # Throttle interval after first run.
            time.sleep(interval)

        try:
            pool = pool or app.postgres.pool()
        except Exception as e:
            logger.error("Failed to connect to Postgres: %s", e)
        else:
            try:
                for attempt in pool.auto_reconnect():
                    with attempt:
                        dashboard_collector_worker(app, pool)
            except Exception as e:
                logger.error("Dashboard collector error: %s", e)

        elapsed = utcnow() - start
        elapsed = elapsed.total_seconds()


class DashboardPlugin:
    PG_MIN_VERSION = (90400, 9.4)
    s = "dashboard"
    option_specs = [
        OptionSpec(s, "scheduler_interval", default=2, validator=int),
        OptionSpec(s, "history_length", default=150, validator=int),
    ]
    del s

    def __init__(self, app, **kw):
        self.app = app
        self.app.config.add_specs(self.option_specs)

    def bootstrap(self):
        db.bootstrap(self.app.config.temboard.home, "dashboard.db")

    def load(self):
        default_app().mount("/dashboard", bottle)
        self.app.worker_pool.add(workers)
        workers.schedule(id="dashboard_collector_batch", redo_interval=BATCH_DURATION)(
            dashboard_collector_batch_worker
        )
        self.app.scheduler.add(workers)



import json
from collections import UserDict
from datetime import datetime, timezone

_UNDEFINED = object()


def dict_factory(iterable=_UNDEFINED, **kw):
    # a dict factory that do not copy.
    if iterable is _UNDEFINED:
        return kw
    elif isinstance(iterable, dict):
        return iterable
    else:
        return dict(iterable, **kw)


def strtobool(value):
    if not value:
        return False
    value = str(value).lower()
    if value in ("y", "yes", "t", "true", "on", "1"):
        return True
    if value in ("n", "no", "f", "false", "off", "0"):
        return False
    raise ValueError("invalid truth value %s" % value)


class DotDict(UserDict):
    # A wrapper around dict that allows read and write through dot style
    # accessors.

    def __init__(self, *a, **kw):
        # don't call super()__init__ to avoid RecursionError
        self.__dict__["data"] = dict_factory(*a, **kw)

    def __getattr__(self, name):
        try:
            value = self[name]
            # Lazy recursion of DotDict
            if isinstance(value, dict):
                self[name] = value = DotDict(value)
            return value
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        # only public property are in the dictionnary
        if name.startswith("_"):
            super().__setattr__(name, value)
        else:
            if hasattr(value, "items"):
                value = DotDict(value)
            self[name] = value

    def __setstate__(self, state):
        self.__dict__.update(state)

    def setdefault(self, name, default):
        if hasattr(default, "items"):
            default = DotDict(default)
        return UserDict.setdefault(self, name, default)


def ensure_bytes(value, encoding="utf-8"):
    if not hasattr(value, "lower"):
        value = str(value)
    if hasattr(value, "isdecimal"):
        return value.encode(encoding)
    return value


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        elif isinstance(obj, bytes):
            return obj.decode("utf-8")
        else:
            return super().default(obj)


def utcnow():
    # TZ aware UTC now.
    return datetime.utcnow().replace(tzinfo=timezone.utc)



import logging
import os
import platform
import re

from .command import exec_command
from .tools import to_bytes, which

logger = logging.getLogger(__name__)


class Inventory:
    def __init__(self):
        pass


class SysInfo(Inventory):
    def __init__(self):
        self.os = platform.system()

    def cpu_info(self):
        if self.os == "Linux":
            return self._cpu_info_linux()
        else:
            raise Exception("Unsupported OS.")

    def mem_info(self):
        if self.os == "Linux":
            return self._mem_info_linux()
        else:
            raise Exception("Unsupported OS.")

    def ip_addresses(self):
        if self.os == "Linux":
            return self._ip_addresses_linux()
        else:
            raise Exception("Unsupported OS.")

    def file_systems(self):
        if self.os == "Linux":
            return self._file_systems_linux()
        else:
            raise Exception("Unsupported OS.")

    def find_mount_point(self, path, mount_points):
        if self.os == "Linux":
            return self._find_mount_point_linux(path, mount_points)
        else:
            raise Exception("Unsupported OS.")

    def _cpu_info_linux(self):
        cpus = []
        # TODO: implement for other OSes
        if self.os != "Linux":
            return {}
        current_cpu = {}
        with open("/proc/cpuinfo") as f:
            for line in f.read().split("\n"):
                if ":" not in line:
                    continue
                key, value = (part.strip() for part in line.split(":", 1))
                if key == "processor":
                    current_cpu = {}
                    cpus.append(current_cpu)
                if key in [
                    "core id",
                    "cpu MHz",
                    "model name",
                    "cache size",
                    "processor",
                ]:
                    current_cpu[str(re.sub(r" +", "_", key)).lower()] = value
        if cpus:
            return {"cpus": cpus, "cpu_count": len(cpus)}
        return {}

    def _mem_info_linux(self):
        unit_re = re.compile(r"(\d+) ?(\wB)?")
        mem_values = {}
        with open("/proc/meminfo") as f:
            for line in f.read().split("\n"):
                if ":" not in line:
                    continue
                key, value = (part.strip() for part in line.split(":", 1))
                size, unit = unit_re.match(value).groups()
                size = int(size)
                # convert everything to bytes, if a unit is specified
                if unit:
                    size = to_bytes(size, unit[:-1])
                mem_values[key] = size
        return mem_values

    def _ip_addresses_linux(self):
        """Find the host's IP addresses."""
        addrs = []
        try:
            ip = which("ip")
            (rc, out, err) = exec_command([ip, "addr", "show"])
            if rc == 0:
                for line in out.decode("utf8").splitlines():
                    m = re.match(r"^\s+inet ([\d\.]+)/\d+\s", line)
                    if m:
                        addrs.append(m.group(1))

                    m = re.match(r"^\sinet6 ([\dabcdef\:]+)/\d+ scope global", line)
                    if m:
                        addrs.append(m.group(1))
                return addrs
        except OSError:
            pass

        try:
            ifconfig = which("ifconfig", ["/sbin"])
            (rc, out, err) = exec_command([ifconfig, "-a"])
            if rc == 0:
                for line in out.decode("utf8").splitlines():
                    m = re.match(r"^\s+inet (addr:)?([\d\.]+)\s", line)
                    if m:
                        addrs.append(m.group(2))

                    m = re.match(
                        r"^\sinet6 (addr: )?([\dabcdef\:]+)(/\d+)? " ".+[Gg]lobal$",
                        line,
                    )
                    if m:
                        addrs.append(m.group(2))
                return addrs
        except OSError:
            pass

        return addrs

    def _file_systems_linux(self):
        logger.debug("Inspecting file systems.")
        fs = []
        (rc, out, err) = exec_command([which("df"), "--local", "-k"])
        lines = out.splitlines()
        # Remove header
        del lines[0]
        dev = None
        for line in lines:
            cols = line.decode("utf-8").split()
            # Skip rootfs which is redundant on Debian
            if cols[0] == "rootfs":
                logger.debug("Ignoring rootfs mount point.")
                continue

            # Output of df can be multiline when the name of the
            # device is too large
            if len(cols) in (1, 6):
                dev = cols.pop(0)

            if not cols:
                # Multi-line output, skip to next line for next fields.
                continue

            # cols is now always [total, used, avail, use%, mount_point].
            total, used, _, _, mount_point = cols

            # Skip docker volumes.
            if dev in ("devtmpfs", "overlay", "shm", "tmpfs"):
                logger.debug("Ignoring device %s as %s.", dev, mount_point)
                continue

            if dev.startswith("/dev/loop"):
                logger.debug("Ignoring loopback device %s.", dev)
                continue

            # Skip basic FHS directories.
            _, top_level_dir = mount_point.split("/", 2)[:2]
            if top_level_dir in ("dev", "proc", "run", "sys"):
                logger.debug("Ignoring mount point %s.", mount_point)
                continue

            logger.debug("Found filesystem %s at %s.", dev, mount_point)
            fs.append(
                {
                    "mount_point": mount_point,
                    "device": dev,
                    "total": int(total) * 1024,
                    "used": int(used) * 1024,
                }
            )
            dev = None
        return fs

    def mount_points(self):
        return [fs["mount_point"] for fs in self.file_systems()]

    def _find_mount_point_linux(self, path, mount_points):
        realpath = os.path.realpath(path)

        if not os.path.exists(realpath):
            return None

        # Get the parent dir when it is not a directory
        if not os.path.isdir(realpath):
            realpath = os.path.dirname(realpath)

        # Walk up parents directory
        while True:
            if realpath in mount_points or realpath == "/":
                return realpath

            realpath = os.path.dirname(realpath)


class PgInfo(Inventory):
    def __init__(self, db_conn):
        self.db_conn = db_conn

    def is_in_recovery(self):
        if self.db_conn.server_version >= 90000:
            return self.db_conn.queryscalar("SELECT pg_is_in_recovery() AS standby;")
        return False

    def start_time(self):
        return self.db_conn.queryscalar("SELECT pg_postmaster_start_time();")

    def tablespaces(self, data_directory):
        # Grab the list of tablespaces
        if self.db_conn.server_version < 90200:
            q = """\
            SELECT spcname, spclocation, pg_tablespace_size(oid) AS size
            FROM pg_tablespace;
            """
        else:
            q = """\
            SELECT
              spcname,
              pg_tablespace_location(oid) AS spclocation,
              pg_tablespace_size(oid) AS size
            FROM pg_tablespace;
            """

        tablespaces = []
        sysinfo = SysInfo()
        fs = sysinfo.mount_points()
        for row in self.db_conn.query(q):
            # when spclocation is empty, replace with data_directory
            if row["spclocation"] is None:
                path = data_directory
            else:
                path = row["spclocation"]

            tablespaces.append(
                {
                    "spcname": row["spcname"],
                    "path": path,
                    "mount_point": sysinfo.find_mount_point(path, fs),
                    "size": row["size"],
                }
            )

    def databases(self):
        q = """\
        SELECT
          datname AS dbname, pg_encoding_to_char(encoding) AS encoding,
          pg_database_size(datname) AS size
        FROM pg_database WHERE datallowconn;
        """

        dbs = {}
        for r in self.db_conn.query(q):
            dbs[r["dbname"]] = r
        return dbs



# Managed file-based SQL query for agent
#
# Implement a dict-like interface to loading SQL queries from files.
#
# Implement SQL edition using pragma comment. Line ending with
#
# -- pragma:pg_version_min XXXXXX
#
# or
#
# -- pragma:pg_version_max XXXXXX
#
# Are filtered accodring to instance pg_version (from discovery). See app.py
# for initialization. Version comparison is inclusive. To include a line up to
# a major version, use 149999.
#
import os.path
import re
from textwrap import dedent


class QueryFiler(dict):
    # A dict loading it's values from SQL files in a directory.

    def __init__(self, path):
        self.path = path
        super().__init__()

    def load(self, pg_version=None):
        fmt = dedent("""\
        -- From %s
        %s
        """)

        for path in self.iter_files():
            filename = os.path.basename(path)
            name, ext = os.path.splitext(filename)
            with open(path) as fo:
                if pg_version:
                    sql = "".join(
                        filter(lambda line: filter_pragma_version(line, pg_version), fo)
                    )
                else:
                    sql = fo.read()
            self[name] = fmt % (path, sql)

    def iter_files(self):
        for filename in os.listdir(self.path):
            if not filename.endswith(".sql"):
                continue
            yield self.path + "/" + filename


_pragma_version_re = re.compile(r"-- pragma:pg_version_(min|max) (\d{6})")


def filter_pragma_version(line, pg_version):
    match = _pragma_version_re.search(line)
    if not match:
        return True

    operator = match.group(1)
    target = int(match.group(2))

    if "min" == operator and target <= pg_version:
        return True
    if "max" == operator and pg_version <= target:
        return True

    return False



import logging

from bottle import HTTPError, HTTPResponse, default_app, get, post, request, response

from ..notification import NotificationMgmt
from ..toolkit.signing import InvalidSignature, canonicalize_request, verify_v1

logger = logging.getLogger(__name__)


@get("/error/<error_code:int>", skip=["signature"])
def get_error(error_code):
    if error_code < 400:
        raise HTTPResponse("HTTPResponse raised", error_code)
    elif error_code < 500:
        raise HTTPError(error_code, "HTTPError raised")
    else:
        raise Exception("This is an Exception %s" % error_code)


@get("/discover", skip=["signature"])
def get_discover():
    app = default_app().temboard
    discover = app.discover.ensure_latest().copy()
    discover["signature_status"] = "enabled"

    signature = request.headers.get("x-temboard-signature")
    if signature:
        crequest = canonicalize_request(request.method, request.path, request.headers)
        try:
            if not signature.startswith("v1:"):
                raise InvalidSignature("Unsupported signature version.")
            signature = signature[3:]

            verify_v1(app.config.signing_key, signature, crequest)
            discover["signature_status"] = "valid"
        except InvalidSignature:
            discover["signature_status"] = "invalid"

    response.set_header("ETag", app.discover.etag)
    return discover


@post("/discover")
def post_discover(pgconn):
    app = default_app().temboard
    data = app.discover.refresh(pgconn).copy()
    # POST endpoint does not bypass signature verification.
    data["signature_status"] = "valid"
    response.set_header("ETag", app.discover.etag)
    return data


@get("/notifications")
def notifications():
    config = default_app().temboard.config
    return list(NotificationMgmt.get_last_n(config, -1))


@get("/status", skip=["signature"])
def get_status(pgconn):
    app = default_app().temboard
    response.set_header("X-TemBoard-Discover-ETag", app.discover.etag)
    return app.status.get(pgconn)



"""
The inventory contains support functions to discover the features of the host
(ex: OS, CPU info, MemInfo)

Currently, only Linux is supported.
"""

import logging
import os
import pwd

from ...inventory import PgInfo, SysInfo


def host_info(discover):
    """Gather system information."""
    sinfo = SysInfo()

    s = discover["system"]
    hostinfo = {
        "os": s["os"],
        "os_version": s["os_version"],
        "cpu_arch": s["arch"],
        "hostname": s["fqdn"],
    }
    hostinfo.update(sinfo.cpu_info())
    hostinfo["memory_size"] = s["memory"]
    hostinfo["ip_addresses"] = sinfo.ip_addresses()
    hostinfo["filesystems"] = sinfo.file_systems()
    hostinfo["os_flavor"] = s["distribution"]
    return hostinfo


def instance_info(pool, dbnames, discover):
    """Gather PostgreSQL instance information."""
    p = discover["postgres"]
    conninfo = pool.postgres.pqvars()
    instance_info = {
        "hostname": discover["system"]["fqdn"],
        "instance": p.get("cluster_name"),
        "local_name": p.get("cluster_name"),
        "available": True,
        "host": conninfo["host"],
        "port": conninfo["port"],
        "user": conninfo["user"],
        "database": conninfo["database"],
        "password": conninfo["password"],
    }

    # Try the connection
    try:
        conn = pool.getconn(dbname=conninfo["database"])
    except Exception as e:
        logging.error("Failed to connect to Postgres: %s", e)
        instance_info["available"] = False
        return instance_info

    # Get PostgreSQL informations using PgInfo
    pginfo = PgInfo(conn)
    # Gather the info while where are connected
    instance_info["version_num"] = p["version_num"]
    instance_info["version"] = p["version"]
    instance_info["data_directory"] = p["data_directory"]
    instance_info["max_connections"] = p["max_connections"]
    instance_info["start_time"] = pginfo.start_time()

    # hot standby is available from 9.0
    instance_info["standby"] = pginfo.is_in_recovery()

    # Grab the list of tablespaces
    instance_info["tablespaces"] = pginfo.tablespaces(instance_info["data_directory"])

    # When the user has not given a dbnames list or '*' in the
    # configuration file, we must get the list of databases. Since
    # we have a working connection, let's do it now.
    dbs = pginfo.databases()
    instance_info["dbnames"] = []
    for db in dbnames:
        if db == "*":
            instance_info["dbnames"] = list(dbs.values())
            break
        if db in dbs.keys():
            instance_info["dbnames"].append(dbs[db])

    # Now that we have the data_directory, find the owner
    try:
        statinfo = os.stat(instance_info["data_directory"])
        instance_info["sysuser"] = pwd.getpwuid(statinfo.st_uid).pw_name
    except OSError as e:
        logging.warning("Unable to get the owner of PGDATA: %s", str(e))
        instance_info["sysuser"] = None

    return instance_info



import json
import os
import sqlite3
from textwrap import dedent
from time import time as current_time

from ...toolkit.utils import JSONEncoder


def bootstrap(path, dbname):
    """Create SQLite database model we use to store collected data.

    last_measures table aims to keep a track of the last collected values
    (for some metrics, not all) we need to have to compute delta. This table
    must be purged when the agent starts because we do not want to compute
    delta values with potentially old data resulting with outliers.

    metrics table is used to queued collected data before they are pushed to
    temboard server.
    """

    with sqlite3.connect(os.path.join(path, dbname)) as conn:
        c = conn.cursor()
        c.execute("DROP TABLE IF EXISTS last_measures")
        c.execute(
            dedent("""
                CREATE TABLE last_measures (
                    time REAL,
                    key TEXT PRIMARY KEY,
                    data TEXT
                )
            """)
        )
        c.execute(
            dedent("""
                CREATE TABLE IF NOT EXISTS metrics (
                    time REAL PRIMARY KEY,
                    data TEXT
                )
            """)
        )


def add_metric(path, dbname, time, data):
    with sqlite3.connect(os.path.join(path, dbname)) as conn:
        c = conn.cursor()
        c.execute(
            "INSERT INTO metrics VALUES(?, ?)",
            (time, json.dumps(data, cls=JSONEncoder)),
        )
        # When data are pulled from temboard server, we need to keep 6 hours of
        # data history for recovery.
        time_limit = current_time() - (60 * 60 * 6)
        c.execute("DELETE FROM metrics WHERE time < ?", (time_limit,))


def delete_metric(path, dbname, time):
    with sqlite3.connect(os.path.join(path, dbname)) as conn:
        c = conn.cursor()
        c.execute("DELETE FROM metrics WHERE time = ?", (time,))


def get_metrics(path, dbname, limit=50, start_timestamp=None):
    query = "SELECT time, data FROM metrics"
    args = ()
    if start_timestamp:
        query += " WHERE time >= ?"
        args += (start_timestamp,)
    else:
        # By default we want only the most recent record. This could be
        # achieved in a most elegant an simple way, but we want to keep the
        # same logic wheter or not start_timestamp is in use.
        query += " WHERE time >= (SELECT MAX(time) FROM metrics)"

    query += " ORDER BY time ASC"

    if limit:
        query += " LIMIT ?"
        args += (limit,)

    with sqlite3.connect(os.path.join(path, dbname)) as conn:
        c = conn.cursor()
        c.execute(query, args)
        return c.fetchall()


def get_last_measure(path, dbname, key):
    with sqlite3.connect(os.path.join(path, dbname)) as conn:
        c = conn.cursor()
        c.execute("SELECT time, data FROM last_measures WHERE key = ?", (key,))
        return c.fetchone()


def upsert_last_measure(path, dbname, time, key, data):
    with sqlite3.connect(os.path.join(path, dbname)) as conn:
        c = conn.cursor()
        try:
            c.execute(
                "INSERT INTO last_measures VALUES(?, ?, ?)",
                (time, key, json.dumps(data, cls=JSONEncoder)),
            )
        except sqlite3.IntegrityError:
            c.execute(
                "UPDATE last_measures SET time = ?, data = ? " "WHERE key = ?",
                (time, json.dumps(data, cls=JSONEncoder), key),
            )


def drop_current_for_delta_metrics(metrics):
    # Drop current value. Keeping only delta value.
    for probe, samples in metrics["data"].items():
        samples[:] = [
            sample
            for sample in samples
            # Keep non-delta sample or delta with last measure.
            if "current" not in sample or "measure_interval" in sample
        ]

        for sample in samples:
            sample.pop("current", None)

    return metrics


def use_current_for_delta_metrics(metrics):
    # Override delta measures with current value.
    for probe, samples in metrics["data"].items():
        for sample in samples:
            sample.update(sample.pop("current", {}))
            sample.pop("measure_interval", None)
    return metrics



import logging

from bottle import default_app, request

from ...notification import Notification, NotificationMgmt
from ...queries import QUERIES
from ...tools import validate_parameters
from ...web.app import CustomBottle
from . import functions as activity_functions

bottle = CustomBottle()
logger = logging.getLogger(__name__)


@bottle.get("/")
def get_activity(pgconn):
    limit = int(request.query.get("limit", 300))
    return activity_functions.get_activity(pgconn, limit)


@bottle.get("/waiting")
def get_activity_waiting(pgconn):
    limit = int(request.query.get("limit", 300))
    return activity_functions.get_activity_waiting(pgconn, limit)


@bottle.get("/blocking")
def get_activity_blocking(pgconn):
    limit = int(request.query.get("limit", 300))
    return activity_functions.get_activity_blocking(pgconn, limit)


@bottle.post("/kill")
def post_activity_kill(pgconn):
    app = default_app().temboard
    validate_parameters(request.json, [("pids", int, True)])
    backends = []
    for pid in request.json["pids"]:
        killed = pgconn.queryscalar("SELECT pg_terminate_backend(%s) AS killed" % (pid))
        backends.append(dict(pid=pid, killed=killed))

        # Push a notification.
        try:
            NotificationMgmt.push(
                app.config,
                Notification(
                    username=request.username, message="Backend %s terminated" % (pid)
                ),
            )
        except Exception:
            logger.exception("Failed to push notification:")

    return dict(backends=backends)


LOCKS_COLUMNS = [
    "pid",
    "locktype",
    "database",
    "schema",
    "relation",
    "mode",
    "granted",
    "blocking_pid",
    "waitstart",
    "waiting_pids",
]


@bottle.get("/locks")
def get_locks(pgconn):
    # Returns waiting and blocking locks

    discover = default_app().temboard.discover.ensure_latest()
    if discover["postgres"]["version_num"] < 100000:
        try:
            LOCKS_COLUMNS.remove("waitstart")
        except ValueError:
            pass

    limit = int(request.query.get("limit", 300))
    query = QUERIES["activity-locks"] + " LIMIT %d" % limit

    return dict(columns=SESSION_COLUMNS, rows=[row for row in pgconn.query(query)])


SESSION_COLUMNS = [
    "pid",
    "client_addr",
    "username",
    "application_name",
    "database",
    "state",
    "backend_start",
    "query_start",
    "duration",
    "query",
    "waiting",
    "blocking",
]


@bottle.get("/sessions")
def get_sessions(pgconn):
    # List client session with blocking informations.

    limit = int(request.query.get("limit", 300))
    query = QUERIES["activity-sessions"] + " LIMIT %d" % limit
    rows = []
    for row in pgconn.query(query):
        try:
            data = read_proc_info(row["pid"])
            row["cpu_time"] = sum(
                (data["utime"], data["stime"], data["cutime"], data["cstime"])
            )
            row["memory"] = data["VmRSS"]
            row["proc_state"] = data["state"]
            row["proc_read"] = data["read_bytes"]
            row["proc_write"] = data["write_bytes"]
        except OSError as e:
            logger.debug("Failed to read /proc/%s info: %s.", row["pid"], e)
            row.update(
                dict(
                    cpu_time=None,
                    memory=None,
                    proc_read=None,
                    proc_state=None,
                    proc_write=None,
                )
            )

        rows.append(row)

    return dict(columns=SESSION_COLUMNS, rows=rows)


def read_proc_info(pid):
    procdir = "/proc/%d" % pid
    data = dict()
    with open(procdir + "/status") as fo:
        data.update(parse_proc_human(fo, "VmRSS"))

    with open(procdir + "/stat") as fo:
        data.update(parse_proc_stat(fo.read()))
    with open(procdir + "/io") as fo:
        data.update(parse_proc_human(fo, "read_bytes", "write_bytes"))
    return data


def parse_proc_stat(raw):
    values = raw.split()
    return dict(
        state=values[2],
        utime=int(values[13]),
        stime=int(values[14]),
        cutime=int(values[15]),
        cstime=int(values[16]),
    )


def parse_proc_human(fo, *keys):
    # Parse human readable proc format
    data = dict()
    for line in fo:
        key, value = line.split(":")
        if keys and key not in keys:
            continue
        value = value.strip()
        if value.endswith(" kB"):
            value = int(value[:-3]) * 1024
        try:
            value = int(value)
        except ValueError:
            pass
        data[key] = value
    return data


class ActivityPlugin:
    PG_MIN_VERSION = (90400, 9.4)

    def __init__(self, app, **kw):
        self.app = app

    def load(self):
        default_app().mount("/activity", bottle)



# Simple live performance counters in logfmt.

import logging
import os
import signal

from . import logfmt

logger = logging.getLogger(__name__)
SC_CLK_TCK = os.sysconf("SC_CLK_TCK")


class PerfCounters(dict):
    @classmethod
    def setup(cls, **defaults):
        if "PERF" not in os.environ:
            return
        return cls(**defaults)

    def __init__(self, **defaults):
        defaults = dict(
            io_rchar=0,
            io_wchar=0,
            stime=0,
            utime=0,
            vsize=0,
            pid=os.getpid(),
            load1=0,
            load5=0,
            load15=0,
            **defaults,
        )
        super().__init__(**defaults)

        try:
            self.delay = int(os.environ["PERF"])
        except (KeyError, ValueError):
            self.delay = 60

        self.delay = max(15, self.delay)
        logger.debug("Scheduling perf counters each %s seconds.", self.delay)

    def __str__(self):
        return logfmt.format(**self)

    def run(self):
        try:
            self.snapshot()
        except Exception as e:
            logger.error("Failed to read performances metrics: %s", e)
            return
        logger.debug("%s", self)
        self.schedule()

    def sigalrm_handler(self, *a):
        self.run()

    def schedule(self):
        signal.alarm(self.delay)

    def snapshot(self):
        # I/O
        needles = ["rchar", "wchar"]
        with open("/proc/self/io") as fo:
            for k, v in parse(fo):
                if k not in needles:
                    continue
                self["io_%s" % k] = v
                if not needles:
                    break

        # PROCESS
        with open("/proc/self/stat") as fo:
            stat = fo.read()

        # Fields are documented in proc(5).
        fields = stat.split()

        utime = int(fields[13]) / SC_CLK_TCK
        self["utime"] = utime

        stime = int(fields[14]) / SC_CLK_TCK
        self["stime"] = stime

        vsize = int(fields[22])
        self["vsize"] = vsize

        # GLOBAL LOAD AVG
        with open("/proc/loadavg") as fo:
            loadavg = fo.read()
        self["load1"], self["load5"], self["load15"], _ = loadavg.split(" ", 3)


def parse(lines):
    for line in lines:
        try:
            k, v = line.split(": ")
        except ValueError:
            continue
        yield k, v.strip()



import errno
import logging
import os
import re
from contextlib import contextmanager
from datetime import datetime
from time import gmtime, strftime

from bottle import HTTPError

logger = logging.getLogger(__name__)


def validate_parameters(values, types):
    """
    Verify that each value of dict 'values' is valid. For doing that, we have
    to loop over all 'types' elements which are tuples: ('values' key,
    validation regexp, if the value item currently checked is a list of thing
    to check).
    If values[key] (or each element of it when it's a list) does not match
    with the regexp then we trow an error.
    """
    for key, typ, is_list in types:
        try:
            if isinstance(typ, bytes):
                typ = str(typ.decode("utf-8"))
            if not is_list:
                # If 'typ' is a string, it must be considered as a regexp
                # pattern.
                if isinstance(typ, str) and re.match(typ, str(values[key])) is None:
                    raise HTTPError(406, "Parameter '%s' is malformed." % (key))
                if not isinstance(typ, str) and isinstance(values[key], type(typ)):
                    raise HTTPError(406, "Parameter '%s' is malformed." % (key))
            if is_list:
                for value in values[key]:
                    if isinstance(typ, str) and not re.match(typ, str(value)):
                        raise HTTPError(406, "Parameter '%s' is malformed." % (key))
                    if not isinstance(typ, str) and typ != type(value):
                        raise HTTPError(406, "Parameter '%s' is malformed." % (key))
        except HTTPError as e:
            raise e
        except KeyError:
            raise HTTPError(406, "Parameter '%s' not sent." % (key))
        except Exception as e:
            logger.exception(str(e))
            raise HTTPError(500, "Internal error.")


MULTIPLIERS = ["", "k", "M", "G", "T", "P"]


def to_bytes(size, unit):
    """
    Convert the given size, expressed in unit, to bytes.

    Args:
        size (int): the size to convert to bytes
        unit (str): the unit this size is in
    Returns:
        the converted size

    Example:

    >>> to_bytes(7890, "M")
    8273264640
    >>> to_bytes(7890, "unexistent_unit")
    Traceback (most recent call last):
    ...
    KeyError: 'Invalid unit: unexistent_unit'

    """
    if unit not in MULTIPLIERS:
        raise KeyError("Invalid unit: %s" % unit)
    return size * 1024 ** MULTIPLIERS.index(unit)


def which(prog, search_path=None):
    """
    Search for a file in the given list of directories or fallback to PATH.

    Args:
        prog (str): the file to search
        search_path (list): list of directories to search
    Returns:
        the first path found

    Example:

    >>> which("ls")  # doctest: +SKIP
    '/bin/ls'
    >>> which("ifconfig")  # doctest: +SKIP
    Traceback (most recent call last):
    ...
    OSError: No such file or directory
    >>> which("ifconfig", ["/sbin"])  # doctest: +SKIP
    '/sbin/ifconfig'
    """

    env_path = re.split(r":", os.environ["PATH"])
    if search_path is None:
        search_path = env_path
    else:
        search_path += env_path

    for d in search_path:
        path = d + "/" + prog
        if os.path.exists(path):
            return path

    raise OSError(os.strerror(errno.ENOENT))


def now():
    """Give the current date and time at GMT, suitable for PostgreSQL."""
    return strftime("%Y-%m-%d %H:%M:%S +0000", gmtime())


def fromisoformat(datestr):
    datetime_part, tz_part = datestr.split("+")
    datestr = datetime_part + "+" + tz_part.replace(":", "")
    return datetime.strptime(datestr, "%Y-%m-%dT%H:%M:%S.%f%z")


@contextmanager
def noop_manager(ret=None):
    yield ret



# Pure Python procfs-based process title edition
#
#
# This module implements fancy process title in systemctl status or ps output.
# It replaces: `/usr/bin/python3 -m temboardui.main -c
# /etc/temboard/temboard.conf` with `temboard: scheduler` and so on.
#
# There is not a single portable way accross OS, Python version and even ps
# implementation. The simplest and most portable implemention is argv
# overwriting.
#
# CPython API hides original argv. We need to workaround this by searching
# original argv in memory before overwriting this.
#
# Each process is initialized with a memory segment called [stack] containing,
# in order : argv, argc, environ. /proc/self/maps contains the address of the
# stack. This module searches the address of argv in the stack. Once we have
# original argv address and size, we can overwrite the memory with the new
# title.
#
# We truncate title to not overflow argv memory segment. This mean that the
# title maximum length depends on the number of arguments passed to the
# process.
#
# We pad title with null to fill the argv memory segment. This is needed to
# clean end of the title to be displayed in ps output. Otherwise, you'll see
# `temboard: mainon -c /etc/temboard/temboard.conf` instead of `temboard:
# main`, or garbaged from memory.

import ctypes
import logging
import os
import sys

logger = logging.getLogger(__name__)
_setter = None


def set(title):
    return _setter(title)


def init(prefix):
    global _setter
    assert not _setter
    _setter = ProcTitleSetter(prefix)
    _setter.setup()
    set("init")


class ProcTitleSetter:
    def __init__(self, prefix):
        self.prefix = prefix
        # Memory segment containing argv. Will be overriden with new title.
        self.address = None
        # Size of the memory segment containing argv. Will be used to truncate
        # or pad title.
        self.size = None

    def setup(self):  # pragma: nocover
        try:
            # Get argv as defined by the kernel/libc when executing the
            # process. sys.argv does not contains the full argv.
            argv = get_argv_from_pythonapi()
            # Find process stack from /proc/self/maps.
            with open("/proc/self/maps") as fo:
                start, length = find_stack_segment_from_maps(fo)
            argv, self.address = find_argv_in_stack(start, length, argv)
            self.size = carray_size(argv)
            logger.debug("argv is at %#x, len=%d.", self.address, self.size)
        except Exception as e:
            self.address = self.size = False
            logger.debug("Failed to find argv memory segment: %s", e)

    def settitle(self, title):
        if not self.address:
            return

        # cf. https://dali.bo/tYIql
        title = self.prefix + str(title)
        title = title.encode("utf-8")
        # Truncate title to fit in argv memory segment.
        title = title[: self.size - 1]
        # Pad title with NULL bytes to fill the argv segment.
        title = title.ljust(self.size, b"\0")
        # Overwrite argv segment with proc title
        src = ctypes.create_string_buffer(title)
        ctypes.memmove(self.address, src, self.size)
        return title

    def __call__(self, title):
        return self.settitle(title)


def get_argv_from_pythonapi():  # pragma: nocover
    # Wraps CPython API call using ctypes.

    # Allocate variable to point to argv
    arga = ctypes.POINTER(ctypes.c_wchar_p)()
    argc = ctypes.c_int()

    # Get them from CPython API.
    ctypes.pythonapi.Py_GetArgcArgv(ctypes.byref(argc), ctypes.byref(arga))

    # Transform C Array as Python list.
    argv = [arga[i] for i in range(argc.value)]
    return fix_argv(argv)


def find_argv_in_stack(start, length, argv, environ=os.environ):
    # Given an argv list, find the memory segment containing it in
    # /proc/self/maps. This is done by walking backward the stack segment until
    # argv[0] is found. argv[0] is found once the first environment variable is
    # found.
    #
    # stack segment is contains in order argv, argc, environ, and some
    # data irrelevant here.

    argv = argv[:]
    remaining_argv = argv[:]
    found_environ = False
    stackend = start + length - 1
    strings = reverse_find_nulstring(stackend, length)
    for address, string in strings:
        # Check if string is an entry from environ list.
        name, _, _ = string.partition("=")
        if name in environ:
            found_environ = True
            continue
        elif not found_environ:
            # We are after environ, just skip until we find an env var
            # declaration.
            continue
        elif string == remaining_argv[-1]:
            remaining_argv.pop()
            if not remaining_argv:
                # Ok we found all args!
                return argv, address
        elif remaining_argv[-1] == "__COMMAND_STRING__":
            logger.debug("Actual command string: %r.", string)
            remaining_argv.pop()
            command_index = len(remaining_argv)
            argv[command_index] = string
        else:
            raise Exception("Can't find argv in stack segment")


def find_stack_segment_from_maps(lines):
    # See proc(3) for a description of /proc/self/maps format.
    # [stack] is a pseudo path to initial stack segment.
    for line in lines:
        if not line.endswith("[stack]\n"):
            continue
        segment = line.split()
        address_range = segment[0]
        start, end = address_range.split("-")
        start = int(start, base=16)
        end = int(end, base=16)
        return start, end - start
    else:
        raise Exception("Can't find stack segment")


def compute_main_module_name(mod):
    # Fix __main__ module to find its importable name.

    if mod.__name__ == "__main__":
        dir_, file_ = mod.__file__.rsplit("/", 1)
        name = file_.replace(".py", "")
    else:
        name = mod.__name__

    return mod.__package__ + "." + name if mod.__package__ else name


def fix_argv(argv):
    # CPython alters argv. This breaks searching for argv in memory. This
    # function reverses CPython alteration to libc original argv.
    #
    # Alterations are: adding a -c argument, trimming of main module name.
    #
    # Replaces `-c, -c` with `-c __COMMAND_STRING__`.
    # Replaces `-m -m ` with `-m <effective.main.module>`

    modname = None
    # Whether the current -c is placeholder for command string
    command_string = False
    for i, arg in enumerate(argv[:]):
        if 0 == i:
            # Python interpreter. Skip it.
            continue
        elif not arg.startswith("-"):
            # Python argument. Next argv items are scripts arguments. Stop now.
            break
        elif "-" == arg:
            break
        elif "-c" == arg:
            if command_string:
                # Found the second -c, replace it with a placeholder.
                argv[i] = "__COMMAND_STRING__"
                break
            else:
                # Next -c will be a placeholder for actual python online passed
                # as argument to Python
                command_string = True
        elif "-m" == arg and modname is None:
            # Restore main module name.
            modname = compute_main_module_name(sys.modules["__main__"])
            # In PY3, -m module is replaced with -m -m.
            argv[i + 1] = modname
            break

    return argv


def reverse_find_nulstring(address, limit=8192):
    # Search backward null-terminated strings.
    string = ""
    null = 0
    for addr, b in reverse_walk_memory(address, limit):
        if null == b:
            if string:
                yield addr + 1, string
            string = ""
        else:
            string = chr(b) + string


def reverse_walk_memory(address, limit):
    # yields bytes of a memory segment, one by one, in reverse order.
    pos = address
    for i in range(limit):
        pos = address - i
        yield pos, read_byte(pos)


def read_byte(address):
    type_ = ctypes.POINTER(ctypes.c_char)
    return ctypes.cast(address, type_).contents.value[0]


def carray_size(list_):
    return sum(len(s) for s in list_) + len(list_)


def test_main():  # pragma: nocover
    # Test proctitle editing. Set WAIT=1 environ for the process to wait for
    # RET before failing.

    logging.basicConfig(level=logging.DEBUG)
    setproctitle = ProcTitleSetter(prefix="temboard-process: ")
    setproctitle.setup()
    setproctitle("test process")
    with open("/proc/self/cmdline") as fo:
        cmdline = fo.read()
    logger.debug("/proc/self/cmdline is %r", cmdline)
    wanted = "temboard-process: test process"
    if wanted in cmdline:
        logger.info("PASS")
        return 0

    if os.environ.get("WAIT"):
        logger.debug("Process title should be %r", wanted)
        print("Hit RET to terminate.", file=sys.stderr, end="")
        input()

    logger.error("FAIL")
    return 1


if "__main__" == __name__:  # pragma: nocover
    os._exit(test_main())



import logging
import os

from ..plugins.monitoring import db
from ..toolkit import services
from ..toolkit.app import SubCommand
from .app import app

logger = logging.getLogger(__name__)


@app.command
class Serve(SubCommand):
    """Combined web and worker services."""

    is_service = True

    def main(self, args):
        # Purge all legacy data queues
        home = self.app.config.temboard.home
        if os.path.exists(home):
            [
                os.remove(os.path.join(home, f))
                for f in os.listdir(home)
                if f.endswith(".q")
            ]

        if "monitoring" in self.app.config.temboard.plugins:
            logger.info("Resetting monitoring data.")
            db.bootstrap(self.app.config.temboard.home, "monitoring.db")

        self.app.config.load_signing_key()
        self.app.discover.refresh()
        self.app.discover.write()

        return services.run(self.app.httpd, self.app.scheduler, self.app.worker_pool)



import datetime
import getpass
import logging
import os
from argparse import _VersionAction
from socket import getfqdn
from textwrap import dedent

try:
    import hupper
except ImportError:
    hupper = None

from ..core import workers
from ..discover import Discover, inspect_versions
from ..notification import NotificationMgmt
from ..postgres import Postgres
from ..queries import QUERIES
from ..status import Status
from ..toolkit import proctitle, taskmanager
from ..toolkit import validators as v
from ..toolkit.app import BaseApplication, define_core_arguments
from ..toolkit.configuration import MergedConfiguration, OptionSpec
from ..toolkit.errors import UserError
from ..toolkit.signing import load_public_key
from ..toolkit.tasklist.sqlite3_engine import TaskListSQLite3Engine
from ..version import __version__
from ..web import HTTPDService

logger = logging.getLogger("temboardagent.scripts.agent")


class TemboardAgentApplication(BaseApplication):
    PROGRAM = "temboard-agent"
    VERSION = __version__

    DEFAULT_CONFIGFILES = [
        "/etc/temboard-agent/temboard-agent.conf",
        "temboard-agent.conf",
    ]
    DEFAULT_PLUGINS = [
        "activity",
        "administration",
        "dashboard",
        "maintenance",
        "monitoring",
        "pgconf",
        "statements",
    ]

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.config = TemboardAgentConfiguration()

    def main(self, argv, environ):
        parser = self.create_parser(
            description=dedent("""\
            temBoard agent %s.

            COMMAND is optional. Default command is `serve`, the combined
            service. See available commands below.

            """)
            % __version__
        )
        self.define_arguments(parser)
        args = parser.parse_args(argv)

        command_name = getattr(args, "command_fullname", "serve")
        command = self.commands[command_name]

        self.httpd = HTTPDService(self)

        task_queue = taskmanager.Queue()
        event_queue = taskmanager.Queue()

        self.worker_pool = taskmanager.WorkerPoolService(
            app=self, task_queue=task_queue, event_queue=event_queue
        )
        self.services.append(self.worker_pool)
        self.worker_pool.add(workers)

        self.scheduler = taskmanager.SchedulerService(
            app=self, task_queue=task_queue, event_queue=event_queue
        )
        self.services.append(self.scheduler)

        self.bootstrap(args=args, environ=environ, service=command.is_service)
        self.log_versions()
        config = self.config

        self.discover = Discover(self)
        self.discover.read()
        self.status = Status(self)

        # TaskList engine setup must be done before we load the plugins
        self.scheduler.task_list_engine = TaskListSQLite3Engine(
            os.path.join(config.temboard["home"], "agent_tasks.db")
        )

        self.apply_config()

        if self.debug and hupper and command.is_service:
            self.setup_autoreload()

        if "." not in self.config.temboard.hostname:
            logger.warning("Hostname %s is not a FQDN.", self.config.temboard.hostname)

        procprefix = "%s: " % self.PROGRAM
        cluster_name = self.discover.data["postgres"].get("cluster_name")
        if cluster_name:
            procprefix += cluster_name + ": "
        proctitle.init(procprefix)

        QUERIES.load(self.discover.data["postgres"].get("version_num"))

        self.start_datetime = datetime.datetime.now()
        self.reload_datetime = None
        self.pid = os.getpid()
        self.user = getpass.getuser()

        self.bootstrap_plugins()

        # Boostraping action logs table
        NotificationMgmt.bootstrap(config)

        return command.main(args)

    def apply_config(self):
        self.postgres = Postgres(app=self, **self.config.postgresql)
        self.postgres.connection_lost_observers.append(self.discover.connection_lost)
        return super().apply_config()

    def bootstrap_plugins(self):
        for plugin_name, plugin in self.plugins.items():
            if hasattr(plugin, "bootstrap"):
                logger.debug("Boostraping plugin %s", plugin_name)
                plugin.bootstrap()

    def check_compatibility(self, pg_version):
        # check for compatibility with plugins
        for name, plugin in self.plugins.items():
            if pg_version < plugin.PG_MIN_VERSION[0]:
                logger.error(
                    "%s plugin is incompatible with Postgres below %s",
                    name,
                    plugin.PG_MIN_VERSION[1],
                )

    def define_arguments(self, parser):
        define_core_arguments(parser)
        parser.add_argument(
            "-V", "--version", action=VersionAction, help="show version and exit"
        )
        super().define_arguments(parser)

    def init_specs(self, app_specs):
        specs = super().init_specs(app_specs)

        def add_specs(*new_specs):
            for spec in new_specs:
                specs[str(spec)] = spec

        # These are *core* because they are needed to load plugins.
        s = "postgresql"
        add_specs(
            OptionSpec(s, "host", default="/var/run/postgresql", validator=v.dir_),
            OptionSpec(s, "port", default=5432, validator=v.port),
            OptionSpec(s, "user", default="postgres"),
            OptionSpec(s, "password"),
            OptionSpec(s, "dbname", default="postgres"),
        )

        return specs

    def core_specs(self):
        for spec in super().core_specs():
            yield spec

        for name, spec in self.config_specs.items():
            if name.startswith("postgresql_"):
                yield spec

    def log_versions(self):
        versions = inspect_versions()
        logger.debug("Running on %s %s.", versions["distname"], versions["distversion"])
        logger.debug("Using Python %s (%s).", versions["python"], versions["pythonbin"])
        logger.debug(
            "Using libpq %s, Psycopg2 %s.", versions["libpq"], versions["psycopg2"]
        )

    def reload(self):
        super().reload()
        self.reload_datetime = datetime.datetime.now()

    def setup_autoreload(self):
        logger.debug("Enabling code autoreload using hupper.")
        reloader = hupper.start_reloader("temboardagent.__main__.main")
        reloader.watch_files(QUERIES.iter_files())
        reloader.watch_files(
            filter(
                None,
                [
                    self.config.temboard.configfile,
                    self.config.temboard.ssl_ca_cert_file,
                    self.config.temboard.ssl_cert_file,
                    self.config.temboard.ssl_key_file,
                ],
            )
        )


class TemboardAgentConfiguration(MergedConfiguration):
    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self._signing_key = None

    @property
    def signing_key(self):
        # Lazy load signing key.
        if not self._signing_key:
            self.load_signing_key()
        return self._signing_key

    def load_signing_key(self):
        path = self.temboard.signing_public_key
        logger.debug("Loading signing key from %s.", path)
        try:
            fo = open(path, "rb")
        except OSError as e:
            raise UserError("Failed to load signing key: %s" % e)

        with fo:
            self._signing_key = load_public_key(fo.read())


class VersionAction(_VersionAction):
    fmt = dedent("""\
    temBoard agent %(temboard)s (%(temboardbin)s)
    System %(distname)s %(distversion)s
    Python %(python)s (%(pythonbin)s)
    bottle %(bottle)s
    cryptography %(cryptography)s
    libpq %(libpq)s
    psycopg2 %(psycopg2)s
    """)

    def __call__(self, parser, *_):
        print((self.fmt % inspect_versions()).strip())
        parser.exit()


def list_options_specs():
    # Generate each option specs.
    section = "temboard"
    yield OptionSpec(section, "ui_url", validator=v.url)
    yield OptionSpec(section, "address", default="0.0.0.0", validator=v.address)
    yield OptionSpec(section, "port", validator=v.port, default=2345)
    yield OptionSpec(
        section, "ssl_cert_file", default=OptionSpec.REQUIRED, validator=v.file_
    )
    yield OptionSpec(
        section, "ssl_key_file", default=OptionSpec.REQUIRED, validator=v.file_
    )
    yield OptionSpec(section, "ssl_ca_cert_file", validator=v.file_)
    yield OptionSpec(
        section, "signing_public_key", default="signing-public.pem", validator=v.path
    )
    yield OptionSpec(section, "hostname", default=getfqdn(), validator=v.fqdn)
    home = os.environ.get("HOME", "/var/lib/temboard-agent")
    yield OptionSpec(section, "home", default=home, validator=v.writeabledir)


app = TemboardAgentApplication(specs=list_options_specs())



# services manages foreground and background services.
#
# run() function is the main entry point and should be enough.
# It handles signals, background services, perf metrics and the main loop.
#
# The core concept is the Service. A service is an object responsible to answer
# requests in a dedicated long running process. temBoard has services for
# Tornado web server, task scheduler, worker pool and plain HTTP server.
#
# An IO loop runs the service. Either tornado asyncio loop or a custom stub
# in syncio.py for synchronous loop.
#
# A key feature for good multiprocessing is proper signal handling.
# BackgroundManager handles propagation of signals and proper behaviour on kill
# of parent of child services.
#
# Since POSIX allows one single signal handler per signal, the
# SignalMultiplexer dispatches signals to idoine method of any object passed to
# it. SignalMultiplexer ask the loop to register signal handler.
#
# To test the behaviour of this code, run temBoard live and:
#
# - Type Ctrl+C, everything should stop gracefully.
# - pkill -ef 'temboard: web' -INT, same.
# - pkill -ef 'temboard: web' -TERM, same.
# - pkill -ef 'temboard: web' -KILL: scheduler and worker pool should stop.
# - pkill -ef 'temboard: scheduler' -KILL: scheduler is restarted
# - pkill -ef 'temboard: web' -HUP : all process must reload configuration
# - Tornado autoreload: all processes are restarted.
#
# Same for temBoard agent.


import logging
import os
import signal
import time

from . import proctitle

logger = logging.getLogger(__name__)


def run(main, *backgrounds):
    """Run a main service object function in this process.

    Handle signals for INT, TERM, HUP, CHLD and ALRM.
    Handle background services.
    Handle perf metrics.
    Start the loop provided by the service.
    Returns an exit code.

    Note that BackgroundManager calls this same function for each background
    service.

    """

    proctitle.set(main)
    loop = main.create_loop()
    sgm = SignalMultiplexer(loop)
    sgm.register(LoopStopper(loop))
    sgm.register(main)

    bg = BackgroundManager(loop)
    if backgrounds:
        for service in backgrounds:
            bg.add(service)
        sgm.register(bg)

    main.setup(sgm, bg)
    with sgm, bg:
        logger.debug("Entering %s loop.", main)
        loop.start()
    if hasattr(main, "teardown"):
        main.teardown()
    logger.debug("Done. service=%s", main)
    return 0


def execute(service):
    """Execute an external command in this process.

    Replace current Python program by a service.command.
    Does not return. service.command continues process life.
    """

    if hasattr(service, "setup"):
        service.setup()

    # Close all files except stdin, stdout, stderr.
    for fd in range(3, os.sysconf("SC_OPEN_MAX")):
        try:
            os.close(fd)
        except OSError:
            pass
    os.execvp(service.command[0], service.command)


class LoopStopper:
    def __init__(self, loop):
        self.loop = loop
        self.stopping = False

    # Interface for SignalMultiplexer
    def sigint_handler(self, *a):
        self.loop.stop()
        logger.info("Interrupted.")
        self.stopping = True

    def sigterm_handler(self, *a):
        if self.stopping:
            # Background service receives both SIGINT from terminal and SIGTERM
            # from main process. Prevent restopping on SIGTERM after SIGINT.
            return
        self.loop.stop()
        logger.info("Terminated.")


class BackgroundManager:
    def __init__(self, loop):
        self.loop = loop
        self.services = {}
        self.pids = {}
        self.stopping = False  # Whether to restart on SIGCHLD

    def add(self, service):
        self.services[str(service)] = service

    def __bool__(self):
        return bool(self.services)

    def __enter__(self):
        if not self.services:
            return

        self._read_pids()

        if self.pids:
            logger.debug("Cleaning previous background services.")
            self.kill()
            if self.wait():
                raise Exception("Background services are still alive.")
        self.start()

    def __exit__(self, etype, evalue, etb):
        if not self.services:
            return
        if etype:
            logger.warning("Exiting on error. err=%s", evalue)
        self.stop()

    def start(self):
        """Start unstarted services."""

        for name, service in self.services.items():
            if name in self.pids:
                continue
            self.fork(service)

    def fork(self, service):
        pid = os.fork()
        if pid:  # Parent process
            logger.debug("Background service started. service=%s pid=%d", service, pid)
            self._save_pid(service, pid)
            return pid

        # Child process
        if hasattr(self.loop, "asyncio_loop"):
            # Cleanup parent signals handling.
            # See https://bugs.python.org/issue22087 and https://bugs.python.org/issue21998 for details about asyncio and fork.
            signal.set_wakeup_fd(-1)

        if hasattr(service, "command"):
            execute(service)
        else:
            os._exit(run(service))

    def _read_pids(self):
        for name, service in self.services.items():
            if not hasattr(service, "pidfile"):
                continue
            if not os.path.exists(service.pidfile):
                continue
            with open(service.pidfile) as fo:
                self.pids[name] = int(fo.read().strip())
                logger.debug(
                    "Read pid from pidfile. service=%s pid=%d", name, self.pids[name]
                )

    def _save_pid(self, service, pid):
        self.pids[str(service)] = pid
        if hasattr(service, "pidfile"):
            with open(service.pidfile, "w") as fo:
                fo.write(str(pid))

    def _drop_pid(self, name):
        del self.pids[name]
        s = self.services[name]
        if not hasattr(s, "pidfile"):
            return
        if os.path.exists(s.pidfile):
            os.unlink(s.pidfile)

    def stop(self):
        self.stopping = True
        for name, pid in self.pids.items():
            logger.debug("Terminating background service. service=%s pid=%d", name, pid)
            try:
                os.kill(pid, signal.SIGTERM)
            except ProcessLookupError:
                pass
        time.sleep(0.125)
        if self.wait():
            self.kill()
            # Wait for children. If we don't wait, we'll have zombies.
            self.wait()

    def wait(self, timeout=5, step=0.5):
        """Return true if process are still alive"""
        if not self.services:
            return False

        logger.debug("Waiting background services.")
        while timeout > 0 and self.pids:
            for name, pid in self.pids.copy().items():
                try:
                    pid, status = os.waitpid(-1, os.WNOHANG)
                except ChildProcessError:
                    pass
                if pid:
                    self._drop_pid(name)

            time.sleep(step)
            timeout -= step

        return bool(self.pids)

    def kill(self, sig=signal.SIGKILL):
        for name, pid in self.pids.items():
            logger.warning(
                "Signaling background service. service=%s pid=%s signal=%s",
                name,
                pid,
                sig,
            )
            try:
                os.kill(pid, sig)
            except ProcessLookupError:
                pass

    # interface for SignalMultiplexer.
    def sighup_handler(self, *a):
        for pid in self.pids.values():
            os.kill(pid, signal.SIGHUP)

    def sigchld_handler(self, *a):
        if self.stopping:
            return

        for name, service in self.services.items():
            try:
                pid, _ = os.waitpid(self.pids[name], os.WNOHANG)
                if pid == 0:  # Still alive
                    continue
            except ChildProcessError:
                pass

            logger.warning(
                "Background service dead. Restarting. service=%s pid=%s", name, pid
            )
            self._drop_pid(name)

        self.start()

    def sigint_handler(self, *_):
        # If background services got SIGINT too, don't restart them. If only
        # main service received SIGINT, let __exit__ call effective stop().
        self.stopping = True


class SignalMultiplexer:
    def __init__(self, loop):
        # Map signal to handler list.
        self.handlers = {}
        if hasattr(loop, "asyncio_loop"):
            loop = loop.asyncio_loop
        self.loop = loop

    def register(self, handler):
        if hasattr(handler, "sigalrm_handler"):
            self.handlers.setdefault(signal.SIGALRM, []).append(handler.sigalrm_handler)
        if hasattr(handler, "sigchld_handler"):
            self.handlers.setdefault(signal.SIGCHLD, []).append(handler.sigchld_handler)
        if hasattr(handler, "sighup_handler"):
            self.handlers.setdefault(signal.SIGHUP, []).append(handler.sighup_handler)
        if hasattr(handler, "sigint_handler"):
            self.handlers.setdefault(signal.SIGINT, []).append(handler.sigint_handler)
        if hasattr(handler, "sigterm_handler"):
            self.handlers.setdefault(signal.SIGTERM, []).append(handler.sigterm_handler)

    def __enter__(self):
        for sig, handlers in self.handlers.items():
            self.loop.add_signal_handler(sig, self._handle, sig)
        return self

    def __exit__(self, *a):
        for sig, handlers in self.handlers.items():
            self.loop.remove_signal_handler(sig)

    def _handle(self, sig, stack_frame=None):
        for handler in self.handlers.get(sig, []):
            handler(sig, stack_frame)



import logging
from datetime import datetime

from bottle import default_app, request

from temboardagent.toolkit import taskmanager
from temboardagent.tools import validate_parameters

from ...web.app import CustomBottle
from . import functions

bottle = CustomBottle()
workers = taskmanager.WorkerSet()
logger = logging.getLogger(__name__)


@bottle.get("/")
def get_instance(pgconn, pgpool):
    instance = next(functions.get_instance(pgconn))

    rows = functions.get_databases(pgconn)

    databases = []
    for database in rows:
        database = dict(database)
        # we need to connect with a different database
        dbname = database["datname"]
        for getconn in pgpool.auto_reconnect():
            with getconn(dbname) as dbconn:
                database.update(**functions.get_database(dbconn))
        databases.append(database)

    return {"instance": instance, "databases": databases}


T_VACUUM_MODE = b"(((^|,)(full|freeze|analyze))+$)"
T_TIMESTAMP_UTC = b"(^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}Z$)"


@bottle.get("/<dbname>")
def get_database(pgpool, dbname):
    for getconn in pgpool.auto_reconnect():
        with getconn(dbname) as conn:
            database = functions.get_database_size(conn)
            schemas = functions.get_schemas(conn)
    return dict(database, **{"schemas": schemas})


@bottle.get("/<dbname>/schema/<schema>")
def get_schema(pgpool, dbname, schema):
    for getconn in pgpool.auto_reconnect():
        with getconn(dbname) as conn:
            tables = functions.get_tables(conn, schema)
            indexes = functions.get_schema_indexes(conn, schema)
            schema = functions.get_schema(conn, schema)
    return dict(dict(tables, **indexes), **schema)


@bottle.get("/<dbname>/schema/<schema>/table/<table>")
def get_table(pgpool, dbname, schema, table):
    for getconn in pgpool.auto_reconnect():
        with getconn(dbname) as conn:
            ret = functions.get_table(conn, schema, table)
            ret.update(**functions.get_table_indexes(conn, schema, table))
    return ret


@bottle.post("/<dbname>/vacuum")
def post_vacuum_database(pgpool, dbname):
    return post_vacuum(pgpool, request.json, dbname)


@bottle.post("/<dbname>/schema/<schema>/table/<table>/vacuum")
def post_vacuum_table(pgpool, dbname, schema, table):
    return post_vacuum(pgpool, request.json, dbname, schema, table)


def post_vacuum(pgpool, post, dbname, schema=None, table=None):
    app = default_app().temboard
    # Parameters format validation
    if "datetime" in post:
        validate_parameters(post, [("datetime", T_TIMESTAMP_UTC, False)])
    dt = post.get("datetime", datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"))
    if "mode" in post:
        validate_parameters(post, [("mode", T_VACUUM_MODE, False)])
    mode = post.get("mode", "")

    for getconn in pgpool.auto_reconnect():
        with getconn(dbname) as conn:
            return functions.schedule_vacuum(conn, dbname, mode, dt, app, schema, table)


@bottle.get("/<dbname>/vacuum/scheduled")
def scheduled_vacuum_database(dbname):
    app = default_app().temboard
    return functions.list_scheduled_vacuum(app, dbname=dbname)


@bottle.get("/<dbname>/schema/<schema>/table/<table>/vacuum/scheduled")
def scheduled_vacuum_table(dbname, schema, table):
    app = default_app().temboard
    return functions.list_scheduled_vacuum(
        app, dbname=dbname, schema=schema, table=table
    )


@bottle.delete("/vacuum/<vacuum_id>")
def delete_vacuum(vacuum_id):
    app = default_app().temboard
    return functions.cancel_scheduled_operation(vacuum_id, app)


@bottle.get("/vacuum/scheduled")
def scheduled_vacuum():
    return functions.list_scheduled_vacuum(default_app().temboard)


@workers.register(pool_size=10)
def vacuum_worker(app, dbname, mode, schema=None, table=None):
    with app.postgres.connect(database=dbname) as conn:
        return functions.vacuum(conn, dbname, mode, schema, table)


@bottle.post("/<dbname>/analyze")
def post_analyze_database(pgpool, dbname):
    return post_analyze(pgpool, request.json, dbname)


@bottle.post("/<dbname>/schema/<schema>/table/<table>/analyze")
def post_analyze_table(pgpool, dbname, schema, table):
    return post_analyze(pgpool, request.json, dbname, schema, table)


def post_analyze(pgpool, post, dbname, schema=None, table=None):
    app = default_app().temboard
    # Parameters format validation
    if "datetime" in post:
        validate_parameters(post, [("datetime", T_TIMESTAMP_UTC, False)])
    dt = post.get("datetime", datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"))
    for getconn in pgpool.auto_reconnect():
        with getconn(dbname) as conn:
            return functions.schedule_analyze(conn, dbname, dt, app, schema, table)


@bottle.get("/<dbname>/analyze/scheduled")
def scheduled_analyze_database(dbname):
    app = default_app().temboard
    return functions.list_scheduled_analyze(app, dbname=dbname)


@bottle.get("/<dbname>/schema/<schema>/table/<table>/analyze/scheduled")
def scheduled_analyze_table(dbname, schema, table):
    app = default_app().temboard
    return functions.list_scheduled_analyze(
        app, dbname=dbname, schema=schema, table=table
    )


@bottle.delete("/analyze/<analyze_id>")
def delete_analyze(analyze_id):
    app = default_app().temboard
    return functions.cancel_scheduled_operation(analyze_id, app)


@bottle.get("/analyze/scheduled")
def scheduled_analyze():
    app = default_app().temboard
    return functions.list_scheduled_analyze(app)


@workers.register(pool_size=10)
def analyze_worker(app, dbname, schema=None, table=None):
    with app.postgres.connect(database=dbname) as conn:
        return functions.analyze(conn, dbname, schema, table)


@bottle.post("/<dbname>/reindex")
def post_reindex_database(pgpool, dbname):
    return post_reindex(pgpool, request.json, dbname)


@bottle.post("/<dbname>/schema/<schema>/table/<table>/reindex")
def post_reindex_table(pgpool, dbname, schema, table):
    return post_reindex(pgpool, request.json, dbname, schema, table)


@bottle.post("/<dbname>/schema/<schema>/index/<index>/reindex")
def post_reindex_index(pgpool, dbname, schema, index):
    return post_reindex(pgpool, request.json, dbname, schema, index=index)


def post_reindex(pgpool, post, dbname, schema=None, table=None, index=None):
    app = default_app().temboard
    # Parameters format validation
    if "datetime" in post:
        validate_parameters(post, [("datetime", T_TIMESTAMP_UTC, False)])
    dt = post.get("datetime", datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"))

    for getconn in pgpool.auto_reconnect():
        with getconn(dbname) as conn:
            return functions.schedule_reindex(
                conn, dbname, dt, app, schema, table, index
            )


@bottle.get("/<dbname>/reindex/scheduled")
def scheduled_reindex_database(dbname):
    app = default_app().temboard
    return functions.list_scheduled_reindex(app, dbname=dbname)


@bottle.get("/<dbname>/schema/<schema>/table/<table>/reindex/scheduled")
def scheduled_reindex_table(dbname, schema, table):
    app = default_app().temboard
    return functions.list_scheduled_reindex(
        app, dbname=dbname, schema=schema, table=table
    )


@bottle.get("/<dbname>/schema/<schema>/reindex/scheduled")
def scheduled_reindex_index(dbname, schema):
    app = default_app().temboard
    return functions.list_scheduled_reindex(app, dbname=dbname, schema=schema)


@bottle.delete("/reindex/<reindex_id>")
def delete_reindex(reindex_id):
    app = default_app().temboard
    return functions.cancel_scheduled_operation(reindex_id, app)


@bottle.get("/reindex/scheduled")
def scheduled_reindex(http_context, app):
    return functions.list_scheduled_reindex(app)


@workers.register(pool_size=10)
def reindex_worker(app, dbname, schema=None, table=None, index=None):
    logger.info(
        "Reindexing. database=%s schema=%s table=%s index=%s",
        dbname,
        schema,
        table,
        index,
    )
    with app.postgres.connect(database=dbname) as conn:
        return functions.reindex(conn, dbname, schema, table, index)


class MaintenancePlugin:
    PG_MIN_VERSION = (90400, 9.4)

    def __init__(self, app, **kw):
        self.app = app

    def load(self):
        default_app().mount("/maintenance", bottle)
        self.app.worker_pool.add(workers)



import json
import os
import re
import time

from ...inventory import SysInfo
from ...notification import NotificationMgmt
from . import db


def get_metrics(app, pool=None):
    res = dict()
    pool = pool or app.postgres.pool()
    for attempt in pool.auto_reconnect():
        with attempt() as conn:
            dm = DashboardMetrics(conn)
            res.update(
                dict(
                    buffers=dm.get_buffers(),
                    hitratio=dm.get_hitratio(),
                    active_backends=dm.get_active_backends(),
                    databases=dm.get_stat_db(),
                )
            )

    dm = DashboardMetrics()
    res.update(
        dict(
            cpu=dm.get_cpu_usage(),
            loadaverage=dm.get_load_average(),
            memory=dm.get_memory_usage(),
            notifications=dm.get_notifications(app.config),
        )
    )

    res.update(dict(timestamp=time.time()))
    return res


def get_metrics_queue(config):
    dm = DashboardMetrics()
    msg = dict()

    row = db.get_last_metric(config.temboard.home, "dashboard.db")
    if row:
        msg = json.loads(row[0])

    msg["notifications"] = dm.get_notifications(config)
    return msg


def get_history_metrics_queue(config):
    return [
        json.loads(d)
        for (d,) in db.get_all_metrics(config.temboard.home, "dashboard.db")
    ]


class DashboardMetrics:
    conn = None
    config = None
    _instance = None

    def __init__(self, conn=None):
        self.conn = conn

    def get_buffers(self):
        return {"nb": self._get_current_buffers()}

    def get_hitratio(self):
        return self.conn.queryscalar("""\
        SELECT CASE sum(blks_hit+blks_read)
          WHEN 0 THEN NULL
          ELSE trunc(sum(blks_hit)/sum(blks_hit+blks_read)*100)::float
        END AS hitratio
        FROM pg_stat_database
        """)

    def get_active_backends(self):
        return {"nb": self._get_current_active_backends()}

    def get_cpu_usage(self):
        sysinfo = SysInfo()
        if sysinfo.os == "Linux":
            return self._get_cpu_usage_linux()

    def get_load_average(self):
        return os.getloadavg()[0]

    def get_memory_usage(self):
        sysinfo = SysInfo()
        if sysinfo.os == "Linux":
            return self._get_memory_usage_linux()

    def get_stat_db(self):
        row = self.conn.queryone("""\
        SELECT
            count(datid) as databases,
            pg_size_pretty(sum(pg_database_size(
                pg_database.datname))::bigint) as total_size,
            sum(pg_database_size(pg_database.datname))::bigint
              AS total_size_bytes,
            sum(xact_commit)::BIGINT as total_commit,
            sum(xact_rollback)::BIGINT as total_rollback
        FROM pg_database
        JOIN pg_stat_database ON (pg_database.oid = pg_stat_database.datid)
        WHERE datistemplate = 'f'
        """)
        return {
            "databases": row["databases"],
            "total_size": row["total_size"],
            "nb": row["databases"],
            "total_size_bytes": row["total_size_bytes"],
            "total_commit": row["total_commit"],
            "total_rollback": row["total_rollback"],
        }

    def _get_memory_usage_linux(self):
        mem_total = 0
        mem_cached = 0
        mem_free = 0
        mem_active = 0
        pattern_line_meminfo = re.compile(r"^([^:]+):\s+([0-9]+) kB$")
        with open("/proc/meminfo") as fd:
            for line in fd.readlines():
                m = pattern_line_meminfo.match(line)
                if m:
                    key = m.group(1)
                    value = m.group(2)
                    if key == "MemTotal":
                        mem_total = int(value)
                    if key == "MemFree":
                        mem_free = int(value)
                    if key == "Cached":
                        mem_cached = int(value)
        if mem_total == 0:
            raise Exception("Can't parse /proc/meminfo.")
        mem_active = mem_total - mem_free - mem_cached
        return {
            "total": mem_total,
            "free": round(float(mem_free) / float(mem_total) * 100, 1),
            "active": round(float(mem_active) / float(mem_total) * 100, 1),
            "cached": round(float(mem_cached) / float(mem_total) * 100, 1),
            "free_bytes": mem_free,
            "active_bytes": mem_active,
            "cached_bytes": mem_cached,
        }

    def _get_cpu_usage_linux(self):
        cpu_time_snap_0 = self._get_current_cpu_usage_linux()
        time.sleep(0.1)
        cpu_time_snap_1 = self._get_current_cpu_usage_linux()
        delta_time_total = 0
        delta = {}
        for k in ["time_user", "time_system", "time_idle", "time_iowait", "time_steal"]:
            delta_time_total += cpu_time_snap_1[k] - cpu_time_snap_0[k]
            delta[k] = cpu_time_snap_1[k] - cpu_time_snap_0[k]
        return {
            "user": 0
            if not delta_time_total
            else round(delta["time_user"] / delta_time_total * 100, 1),
            "system": 0
            if not delta_time_total
            else round(delta["time_system"] / delta_time_total * 100, 1),
            "idle": 0
            if not delta_time_total
            else round(delta["time_idle"] / delta_time_total * 100, 1),
            "iowait": 0
            if not delta_time_total
            else round(delta["time_iowait"] / delta_time_total * 100, 1),
            "steal": 0
            if not delta_time_total
            else round(delta["time_steal"] / delta_time_total * 100, 1),
        }

    def _get_current_cpu_usage_linux(self):
        ret = {}
        with open("/proc/stat") as fd:
            for line in fd.readlines():
                cols = line.split()
                if len(cols) > 0 and cols[0] == "cpu":
                    ret["time_user"] = float(cols[1]) + float(cols[2])
                    ret["time_system"] = (
                        float(cols[3]) + float(cols[6]) + float(cols[7])
                    )
                    ret["time_idle"] = float(cols[4])
                    ret["time_iowait"] = float(cols[5])
                    ret["time_steal"] = float(cols[8])
                    break
        return ret

    def _get_current_buffers(self):
        return self.conn.queryscalar("SELECT buffers_alloc FROM pg_stat_bgwriter")

    def _get_current_active_backends(self):
        if self.conn.server_version >= 90200:
            query = """
SELECT COUNT(*) AS nb FROM pg_stat_activity WHERE state != 'idle'
            """
        else:
            query = """
SELECT COUNT(*) AS nb FROM pg_stat_activity WHERE current_query != '<IDLE>'
            """
        return self.conn.queryscalar(query)

    def get_notifications(self, config):
        return list(NotificationMgmt.get_last_n(config, 15))



import logging

from ..toolkit.app import SubCommand
from .app import app

logger = logging.getLogger(__name__)


@app.command
class RunScript(SubCommand):
    """Run a Python script."""

    def define_arguments(self, parser):
        parser.add_argument("script", metavar="PATH", help="Path to script file.")

        super().define_arguments(parser)

    def main(self, args):
        with open(args.script) as fo:
            source = fo.read()
        code = compile(source, args.script, "exec")

        g = globals()
        g["app"] = self.app

        exec(code, g, g)



# Declare and generates openmetrics sample from monitoring data stored in
# SQLite database.
#
# If possible, implement postgres_exporter and node_exporter metrics. Unlike
# prometheus, temBoard agent compute rates and diff in agent side. The open
# metrics must conform to prometheus exporter way : expose raw data (either
# gauge or counter) and let Prometheus query do computations.

import logging
import sys
from typing import List

from ...tools import fromisoformat

logger = logging.getLogger(__name__)
METADATAS = {}


class MetricMetadata:
    COUNTER = "counter"
    GAUGE = "gauge"

    def __init__(self, name, help_, type_="untyped"):
        self.name = name
        self.help_ = help_
        self.type_ = type_

    def format(self) -> List[str]:
        yield "# HELP {self.name} {self.help_}".format(self=self)
        yield "# TYPE {self.name} {self.type_}".format(self=self)


class Sample:
    def __init__(self, name, labels=None, value=1, timestamp=None):
        self.name = name
        self.labels = labels or {}
        self.value = value
        self.timestamp = timestamp

    def format(self) -> List[str]:
        labels = ",".join(
            [f'{name}="{value}"' for name, value in sorted(self.labels.items())]
        )
        if labels:
            labels = "{%s}" % labels
        if type(self.value) is str:
            raise Exception(f"String value {self.value} for {self.name}.")
        elif type(self.value) in (int, float):
            value = str(self.value)
        elif hasattr(self.value, "timestamp"):
            value = str(self.value.timestamp())
        else:
            raise ValueError(f"Bad value {self.value}")

        tokens = [f"{self.name}{labels}", value, str(self.timestamp or "")]
        yield " ".join(filter(None, tokens))

    def sort_key(self):
        return (self.name, sorted(self.labels.items()))


def format_open_metrics_lines(samples: List[Sample]) -> List[str]:
    described = set()
    for sample in sorted(samples, key=Sample.sort_key):
        if sample.name not in described:
            if described:
                yield ""
            yield from METADATAS[sample.name].format()
            described.add(sample.name)

        yield from sample.format()

    yield ""
    yield "# EOF"
    yield ""


METADATAS.update(
    {
        m.name: m
        for m in [
            MetricMetadata(
                "node_os_info",
                help_=(
                    "Operating system metadata, (see node_uname_info for kernel "
                    "metadata)."
                ),
                type_=MetricMetadata.GAUGE,
            ),
            MetricMetadata(
                "node_uname_info",
                help_=(
                    "Labeled system information as provided by the uname system "
                    "call."
                ),
                type_=MetricMetadata.GAUGE,
            ),
            MetricMetadata(
                "pg_settings_max_connections",
                help_="Sets the maximum number of concurrent connections.",
                type_=MetricMetadata.GAUGE,
            ),
            # For compatibility with postgres_exporter.
            MetricMetadata(
                "pg_exporter_last_scrape_duration_seconds",
                help_="Duration of the last scrape of metrics from PostgresSQL.",
                type_=MetricMetadata.GAUGE,
            ),
            MetricMetadata("pg_static", help_="Version string as reported by postgres"),
            MetricMetadata(
                "pg_up",
                help_=(
                    "Whether the last scrape of metrics from PostgreSQL was able to "
                    "connect to the server (1 for yes, 0 for no)."
                ),
                type_=MetricMetadata.GAUGE,
            ),
            # From
            # https://github.com/prometheus-community/postgres_exporter/blob/master/queries.yaml#L9-L15
            # Required by PostgreSQL Database dashboard variables.
            MetricMetadata(
                "pg_postmaster_start_time_seconds",
                help_="Time at which postmaster started",
                type_=MetricMetadata.GAUGE,
            ),
        ]
    }
)


def generate_samples(temboard_data) -> List[Sample]:
    instance = temboard_data["instances"][0]
    yield Sample(
        "node_os_info", dict(pretty_name=temboard_data["hostinfo"]["os_flavor"])
    )
    nodename, _, domainname = instance["hostname"].partition(".")
    yield Sample(
        "node_uname_info",
        dict(
            domainename=domainname or "(none)",
            machine=temboard_data["hostinfo"]["cpu_arch"],
            nodename=nodename,
            release=temboard_data["hostinfo"]["os_version"],
            sysname=temboard_data["hostinfo"]["os"],
        ),
    )

    yield Sample("pg_up")
    yield Sample("pg_exporter_last_scrape_duration_seconds", value=0.0)
    yield Sample(
        "pg_postmaster_start_time_seconds", value=fromisoformat(instance["start_time"])
    )
    yield Sample("pg_settings_max_connections", value=int(instance["max_connections"]))
    yield Sample(
        "pg_static",
        dict(
            short_version=instance["version"],
            # Extension of postgres_exporter labels.
            temboard_agent_version=temboard_data["version"],
            cluster_name=instance["local_name"],
        ),
    )

    mod = sys.modules[__name__]
    for probe, data in temboard_data["data"].items():
        if not data:
            continue

        generator_name = "generate_%s_samples" % probe
        generator = getattr(mod, generator_name, None)
        if not generator:
            logger.debug("Can't export metrics from probe %s.", probe)
            raise Exception(probe)
            continue

        yield from generator(data, temboard_data["hostinfo"])


METADATAS.update(
    {
        m.name: m
        for m in [
            MetricMetadata(
                "node_load1", help_="1m load average.", type_=MetricMetadata.GAUGE
            ),
            MetricMetadata(
                "node_load5", help_="5m load average.", type_=MetricMetadata.GAUGE
            ),
            MetricMetadata(
                "node_load15", help_="15m load average.", type_=MetricMetadata.GAUGE
            ),
        ]
    }
)


def generate_loadavg_samples(data, hostinfo):
    for relative in "1", "5", "15":
        yield Sample("node_load" + relative, value=float(data[0]["load" + relative]))


METADATAS.update(
    {
        m.name: m
        for m in [
            MetricMetadata(
                "node_cpu_seconds_total",
                help_="Seconds the CPUs spent in each mode.",
                type_=MetricMetadata.COUNTER,
            )
        ]
    }
)


def generate_cpu_samples(data, hostinfo):
    for mode in "idle", "iowait", "steal", "system", "user":
        for core in range(hostinfo["cpu_count"]):
            yield Sample(
                "node_cpu_seconds_total",
                dict(cpu=core, mode=mode),
                # temBoard use global cpu data while node_exporter use per-core
                # data. Divide by core count to mimic per-core data.
                data[0]["time_" + mode] / 1000 / hostinfo["cpu_count"],
            )


METADATAS.update(
    {
        m.name: m
        for m in [
            *[
                MetricMetadata(
                    "node_memory_%s_bytes" % field,
                    help_="Memory information field %s_bytes." % field,
                    type_=MetricMetadata.GAUGE,
                )
                for field in (
                    "Buffers",
                    "Cached",
                    "MemFree",
                    "MemTotal",
                    "SwapTotal",
                    "SwapFree",
                )
            ]
        ]
    }
)


def generate_memory_samples(data, hostinfo):
    yield Sample("node_memory_Buffers_bytes", value=data[0]["mem_buffers"])
    yield Sample("node_memory_Cached_bytes", value=data[0]["mem_cached"])
    yield Sample("node_memory_MemFree_bytes", value=data[0]["mem_free"])
    yield Sample("node_memory_MemTotal_bytes", value=data[0]["mem_total"])
    yield Sample(
        "node_memory_SwapFree_bytes", value=data[0]["swap_total"] - data[0]["swap_used"]
    )
    yield Sample("node_memory_SwapTotal_bytes", value=data[0]["swap_total"])


METADATAS.update(
    {
        m.name: m
        for m in [
            MetricMetadata(
                "node_context_switches_total",
                help_="Total number of context switches.",
                type_=MetricMetadata.COUNTER,
            ),
            MetricMetadata(
                "node_fork_total",
                help_="Total number of forks.",
                type_=MetricMetadata.COUNTER,
            ),
            MetricMetadata(
                "node_procs_blocked",
                help_="Number of processes blocked waiting for I/O to complete.",
                type_=MetricMetadata.GAUGE,
            ),
            MetricMetadata(
                "node_procs_running",
                help_="Number of processes in runnable state.",
                type_=MetricMetadata.GAUGE,
            ),
            MetricMetadata(
                "xnode_procs_total",
                help_="Total number of processes (temBoard custom).",
                type_=MetricMetadata.GAUGE,
            ),
        ]
    }
)


def generate_process_samples(data, hostinfo):
    yield Sample("node_context_switches_total", value=data[0]["context_switches"])
    yield Sample("node_fork_total", value=data[0]["forks"])
    for field in "blocked", "running":
        yield Sample("node_procs_" + field, value=data[0]["procs_" + field])
    yield Sample("xnode_procs_total", value=int(data[0]["procs_total"]))


METADATAS.update(
    {
        m.name: m
        for m in [
            MetricMetadata(
                "node_filesystem_size_bytes",
                help_="Filesystem size in bytes.",
                type_=MetricMetadata.GAUGE,
            ),
            MetricMetadata(
                "node_filesystem_free_bytes",
                help_="Filesystem free space in bytes.",
                type_=MetricMetadata.GAUGE,
            ),
        ]
    }
)


def generate_filesystems_size_samples(data, hostinfo):
    for fs in data:
        yield Sample(
            "node_filesystem_size_bytes",
            dict(device=fs["device"], mountpoint=fs["mount_point"]),
            fs["total"],
        )
        yield Sample(
            "node_filesystem_free_bytes",
            dict(device=fs["device"], mountpoint=fs["mount_point"]),
            fs["total"] - fs["used"],
        )


METADATAS.update(
    {
        m.name: m
        for m in [
            MetricMetadata(
                "pg_locks_count", help_="Number of locks", type_=MetricMetadata.GAUGE
            )
        ]
    }
)


def generate_locks_samples(data, hostinfo):
    lock_modes = (
        "access_share",
        "waiting_access_share",
        "row_share",
        "waiting_row_share",
        "row_exclusive",
        "waiting_row_exclusive",
        "share_update_exclusive",
        "waiting_share_update_exclusive",
        "share",
        "waiting_share",
        "share_row_exclusive",
        "waiting_share_row_exclusive",
        "exclusive",
        "waiting_exclusive",
        "access_exclusive",
        "waiting_access_exclusive",
        "siread",
    )
    for datdata in data:
        for mode in lock_modes:
            prommode = mode.replace("_", "")
            yield Sample(
                "pg_locks_count",
                dict(mode=prommode, datname=datdata["dbname"]),
                datdata[mode],
            )


METADATAS.update(
    {
        m.name: m
        for m in [
            MetricMetadata(
                "pg_stat_activity_count",
                help_="number of connections in this state",
                type_=MetricMetadata.GAUGE,
            )
        ]
    }
)


def generate_sessions_samples(data, hostinfo):
    session_states = {
        "active": "active",
        "disabled": "disabled",
        "fastpath": "fastpath function call",
        "idle": "idle",
        "idle_in_xact": "idle in transaction",
        "idle_in_xact_aborted": "idle in transaction (aborted)",
    }
    for datdata in data:
        for state, promstate in session_states.items():
            yield Sample(
                "pg_stat_activity_count",
                dict(datname=datdata["dbname"], state=promstate),
                datdata[state],
            )


METADATAS.update(
    {
        m.name: m
        for m in [
            MetricMetadata(
                "pg_stat_bgwriter_buffers_alloc",
                help_="Number of buffers allocated",
                type_=MetricMetadata.COUNTER,
            ),
            MetricMetadata(
                "pg_stat_bgwriter_buffers_backend",
                help_="Number of buffers written directly by a backend",
                type_=MetricMetadata.COUNTER,
            ),
            MetricMetadata(
                "pg_stat_bgwriter_buffers_backend_fsync",
                help_=(
                    "Number of times a backend had to execute its own fsync call "
                    "(normally the background writer handles those even when the "
                    "backend does its own write)"
                ),
                type_=MetricMetadata.COUNTER,
            ),
            MetricMetadata(
                "pg_stat_bgwriter_buffers_checkpoint",
                help_="Number of buffers written during checkpoints",
                type_=MetricMetadata.COUNTER,
            ),
            MetricMetadata(
                "pg_stat_bgwriter_buffers_clean",
                help_="Number of buffers written by the background writer",
                type_=MetricMetadata.COUNTER,
            ),
            MetricMetadata(
                "pg_stat_bgwriter_checkpoint_sync_time",
                help_=(
                    "Total amount of time that has been spent in the portion of "
                    "checkpoint processing where files are synchronized to disk, "
                    "in milliseconds"
                ),
                type_=MetricMetadata.COUNTER,
            ),
            MetricMetadata(
                "pg_stat_bgwriter_checkpoint_write_time",
                help_=(
                    "Total amount of time that has been spent in the portion of "
                    "checkpoint processing where files are written to disk, in "
                    "milliseconds"
                ),
                type_=MetricMetadata.COUNTER,
            ),
            MetricMetadata(
                "pg_stat_bgwriter_checkpoints_req",
                help_="Number of requested checkpoints that have been performed",
                type_=MetricMetadata.COUNTER,
            ),
            MetricMetadata(
                "pg_stat_bgwriter_checkpoints_timed",
                help_="Number of scheduled checkpoints that have been performed",
                type_=MetricMetadata.COUNTER,
            ),
            MetricMetadata(
                "pg_stat_bgwriter_maxwritten_clean",
                help_=(
                    "Number of times the background writer stopped a cleaning scan "
                    "because it had written too many buffers"
                ),
                type_=MetricMetadata.COUNTER,
            ),
            MetricMetadata(
                "pg_stat_bgwriter_stats_reset",
                help_="Time at which these statistics were last reset",
                type_=MetricMetadata.COUNTER,
            ),
        ]
    }
)


def generate_bgwriter_samples(data, hostinfo):
    yield Sample(
        "pg_stat_bgwriter_stats_reset", value=fromisoformat(data[0]["stats_reset"])
    )

    bgwriter_stats = (
        "buffers_alloc",
        "buffers_backend",
        "buffers_backend_fsync",
        "buffers_clean",
        "buffers_checkpoint",
        "checkpoint_sync_time",
        "checkpoint_write_time",
        "checkpoints_timed",
        "checkpoints_req",
        "maxwritten_clean",
    )
    for stat in bgwriter_stats:
        if stat not in data[0]:
            continue
        yield Sample("pg_stat_bgwriter_" + stat, value=data[0][stat])


METADATAS.update(
    {
        m.name: m
        for m in [
            MetricMetadata(
                "pg_stat_checkpointer_buffers_written",
                help_="Number of buffers written during checkpoints",
                type_=MetricMetadata.COUNTER,
            ),
            MetricMetadata(
                "pg_stat_checkpointer_sync_time",
                help_=(
                    "Total amount of time that has been spent in the portion of "
                    "checkpoint processing where files are synchronized to disk, "
                    "in milliseconds"
                ),
                type_=MetricMetadata.COUNTER,
            ),
            MetricMetadata(
                "pg_stat_checkpointer_write_time",
                help_=(
                    "Total amount of time that has been spent in the portion of "
                    "checkpoint processing where files are written to disk, in "
                    "milliseconds"
                ),
                type_=MetricMetadata.COUNTER,
            ),
            MetricMetadata(
                "pg_stat_checkpointer_num_requested",
                help_="Number of requested checkpoints that have been performed",
                type_=MetricMetadata.COUNTER,
            ),
            MetricMetadata(
                "pg_stat_checkpointer_num_timed",
                help_="Number of scheduled checkpoints that have been performed",
                type_=MetricMetadata.COUNTER,
            ),
            MetricMetadata(
                "pg_stat_checkpointer_stats_reset",
                help_="Time at which these statistics were last reset",
                type_=MetricMetadata.COUNTER,
            ),
        ]
    }
)


def generate_checkpointer_samples(data, hostinfo):
    yield Sample(
        "pg_stat_checkpointer_stats_reset", value=fromisoformat(data[0]["stats_reset"])
    )

    stats = ("buffers_written", "sync_time", "write_time", "num_timed", "num_requested")
    for stat in stats:
        yield Sample("pg_stat_checkpointer_" + stat, value=data[0][stat])


METADATAS.update(
    {
        m.name: m
        for m in [
            MetricMetadata(
                "xpg_database_size_bytes",
                help_="Database total size",
                type_=MetricMetadata.GAUGE,
            )
        ]
    }
)


def generate_db_size_samples(data, hostinfo):
    for datdata in data:
        yield Sample(
            "xpg_database_size_bytes", dict(datname=datdata["dbname"]), datdata["size"]
        )


METADATAS.update(
    {
        m.name: m
        for m in [
            MetricMetadata(
                "pg_stat_database_temp_bytes",
                help_=(
                    "Total amount of data written to temporary files by queries in "
                    "this database. All temporary files are counted, regardless of "
                    "why the temporary file was created, and regardless of the "
                    "log_temp_files setting.",
                ),
                type_=MetricMetadata.COUNTER,
            )
        ]
    }
)


def generate_temp_files_size_delta_samples(data, hostinfo):
    for datdata in data:
        yield Sample(
            "pg_stat_database_temp_bytes",
            dict(datname=datdata["dbname"]),
            datdata["size"],
        )


METADATAS.update(
    {
        m.name: m
        for m in [
            MetricMetadata(
                "xpg_tablespace_size_bytes",
                help_="Tablespace total size",
                type_=MetricMetadata.GAUGE,
            )
        ]
    }
)


def generate_tblspc_size_samples(data, hostinfo):
    for spcdata in data:
        yield Sample(
            "xpg_tablespace_size_bytes",
            dict(spcname=spcdata["spcname"]),
            spcdata["size"],
        )


METADATAS.update(
    {
        m.name: m
        for m in [
            MetricMetadata(
                "xpg_wal_files_bytes_total",
                help_="Total size of all WALs",
                type_=MetricMetadata.GAUGE,
            ),
            MetricMetadata(
                "xpg_wal_files_total",
                help_="Total number of WALs",
                type_=MetricMetadata.GAUGE,
            ),
            MetricMetadata(
                "xpg_wal_files_archive_ready_total",
                help_="Total number of WALs ready for archiving.",
                type_=MetricMetadata.GAUGE,
            ),
            MetricMetadata(
                "xpg_wal_files_written_size_bytes",
                help_="Total bytes written in current WAL.",
                type_=MetricMetadata.COUNTER,
            ),
        ]
    }
)


def generate_wal_files_samples(data, hostinfo):
    yield Sample("xpg_wal_files_total", value=data[0]["total"])
    yield Sample("xpg_wal_files_archive_ready_total", value=data[0]["archive_ready"])
    yield Sample("xpg_wal_files_bytes_total", value=data[0]["total_size"])
    yield Sample(
        "xpg_wal_files_written_size_bytes",
        dict(current_wal_lsn=data[0]["current_location"]),
        data[0]["written_size"],
    )


METADATAS.update(
    {
        m.name: m
        for m in [
            MetricMetadata(
                "xpg_btree_bloat_ratio",
                help_="Bloat estimation of btree indexes in database.",
                type_=MetricMetadata.GAUGE,
            )
        ]
    }
)


def generate_btree_bloat_samples(data, hostinfo):
    for datdata in data:
        yield Sample(
            "xpg_btree_bloat_ratio", dict(datname=datdata["dbname"]), datdata["ratio"]
        )


METADATAS.update(
    {
        m.name: m
        for m in [
            MetricMetadata(
                "xpg_heap_bloat_ratio",
                help_="Bloat estimation of table data in database.",
                type_=MetricMetadata.GAUGE,
            )
        ]
    }
)


def generate_heap_bloat_samples(data, hostinfo):
    for datdata in data:
        yield Sample(
            "xpg_heap_bloat_ratio", dict(datname=datdata["dbname"]), datdata["ratio"]
        )


METADATAS.update(
    {
        m.name: m
        for m in [
            MetricMetadata(
                "pg_stat_database_xact_commit",
                help_=(
                    "Number of transactions in this database that have been "
                    "committed"
                ),
                type_=MetricMetadata.COUNTER,
            ),
            MetricMetadata(
                "pg_stat_database_xact_rollback",
                help_=(
                    "Number of transactions in this database that have been rolled "
                    "back"
                ),
                type_=MetricMetadata.COUNTER,
            ),
        ]
    }
)


def generate_xacts_samples(data, hostinfo):
    for datdata in data:
        yield Sample(
            "pg_stat_database_xact_commit",
            dict(datname=datdata["dbname"]),
            datdata["n_commit"],
        )
        yield Sample(
            "pg_stat_database_xact_rollback",
            dict(datname=datdata["dbname"]),
            datdata["n_rollback"],
        )


METADATAS.update(
    {
        m.name: m
        for m in [
            MetricMetadata(
                "pg_stat_database_blks_hit",
                help_=(
                    "Number of times disk blocks were found already in the buffer "
                    "cache, so that a read was not necessary (this only includes hits "
                    "in the PostgreSQL buffer cache, not the operating system's file "
                    "system cache)"
                ),
                type_=MetricMetadata.COUNTER,
            ),
            MetricMetadata(
                "pg_stat_database_blks_read",
                help_="Number of disk blocks read in this database",
                type_=MetricMetadata.COUNTER,
            ),
        ]
    }
)


def generate_blocks_samples(data, hostinfo):
    for datdata in data:
        yield Sample(
            "pg_stat_database_blks_hit",
            dict(datname=datdata["dbname"]),
            datdata["blks_hit"],
        )
        yield Sample(
            "pg_stat_database_blks_read",
            dict(datname=datdata["dbname"]),
            datdata["blks_read"],
        )
        # let PromQL compute hitmiss ratio.


METADATAS.update(
    {
        m.name: m
        for m in [
            MetricMetadata(
                "xpg_replication_up",
                help_="Streaming replication connection status",
                type_=MetricMetadata.GAUGE,
            )
        ]
    }
)


def generate_replication_connection_samples(data, hostinfo):
    yield Sample(
        "xpg_replication_up", dict(upstream=data[0]["upstream"]), data[0]["connected"]
    )


METADATAS.update(
    {
        m.name: m
        for m in [
            MetricMetadata(
                "xpg_replication_lag_bytes",
                help_="Streaming replication lag",
                type_=MetricMetadata.GAUGE,
            )
        ]
    }
)


def generate_replication_lag_samples(data, hostinfo):
    yield Sample("xpg_replication_lag_bytes", value=data[0]["lag"])



#
# Configure and extend psycopg2 for temBoard
#
# - Use float for decimal.
# - dict as default row type
# - per query row_factory
# - connection helpers to hide cursor.
#

import logging
import re
from contextlib import closing
from time import sleep

import psycopg2.extensions
from psycopg2 import Error as Psycopg2Error
from psycopg2 import connect
from psycopg2.pool import ThreadedConnectionPool

from .tools import noop_manager

logger = logging.getLogger(__name__)


# See https://www.psycopg.org/docs/faq.html#faq-float
DEC2FLOAT = psycopg2.extensions.new_type(
    psycopg2.extensions.DECIMAL.values,
    "DEC2FLOAT",
    lambda value, curs: float(value) if value is not None else None,
)
psycopg2.extensions.register_type(DEC2FLOAT)


def pg_escape(in_string, escapee_char=r"'"):
    out_string = ""
    out_string += escapee_char
    out_string += re.sub(escapee_char, escapee_char * 2, in_string)
    out_string += escapee_char
    return out_string


class Postgres:
    # main object holding Postgres parameters and methods.

    def __init__(
        self,
        host=None,
        port=5432,
        user=None,
        password=None,
        dbname=None,
        app=None,
        **kw,
    ):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        # Compat with conninfo dict.
        if "database" in kw:
            dbname = kw["database"]
        self.dbname = dbname
        self.app = app
        self._server_version = None
        self.connection_lost_observers = []

    def __repr__(self):
        return "<{} on {}@{}:{}/{}>".format(
            self.__class__.__name__, self.user, self.host, self.port, self.dbname
        )

    def dbpool(self):
        return DBConnectionPool(self)

    def pool(self):
        return ConnectionPool(
            app=self.app,
            observers=self.connection_lost_observers,
            minconn=1,
            maxconn=2,
            **self.pqvars(),
        )

    def connect(self, database=None):
        kw = self.pqvars(database)
        return closing(retry_connect(connect, self.app, **kw))

    def maybe_connect(self, conn=None, database=None):
        return noop_manager(conn) if conn else self.connect(database)

    def copy(self, **kw):
        defaults = dict(
            host=self.host,
            port=self.port,
            user=self.user,
            password=self.password,
            dbname=self.dbname,
            app=self.app,
        )
        kw = dict(defaults, **kw)
        return self.__class__(**kw)

    def pqvars(self, dbname=None):
        return dict(
            host=self.host,
            port=self.port,
            user=self.user,
            password=self.password,
            database=dbname or self.dbname,
            connection_factory=FactoryConnection,
            application_name="temboard-agent",
        )

    def notify_observers(self):
        for observer in self.connection_lost_observers:
            observer()


class FactoryConnection(psycopg2.extensions.connection):  # pragma: nocover
    def cursor(self, *a, **kw):
        row_factory = kw.pop("row_factory", None)
        kw["cursor_factory"] = FactoryCursor.make_factory(row_factory=row_factory)
        return super().cursor(*a, **kw)

    def execute(self, sql, *args):
        with self.cursor() as cur:
            cur.execute(sql, *args)

    def query(self, sql, *args, row_factory=None):
        with self.cursor(row_factory=row_factory) as cur:
            cur.execute(sql, *args)
            yield from cur.fetchall()

    def queryone(self, sql, *args, row_factory=None):
        with self.cursor(row_factory=row_factory) as cur:
            cur.execute(sql, *args)
            return cur.fetchone()

    def queryscalar(self, sql, *args):
        return self.queryone(sql, *args, row_factory=scalar)


class FactoryCursor(psycopg2.extensions.cursor):  # pragma: nocover
    # Implement row_factory for psycopg2.

    @classmethod
    def make_factory(cls, row_factory=None):
        # Build a cursor_factory for psycopg2 connection.
        def factory(*a, **kw):
            kw["row_factory"] = row_factory
            return cls(*a, **kw)

        return factory

    def __init__(self, conn, name=None, row_factory=None):
        super().__init__(conn)
        if not row_factory:

            def row_factory(**kw):
                return kw

        self._row_factory = row_factory

    def fetchone(self):
        row = super().fetchone()
        kw = dict(zip([c.name for c in self.description], row))
        return self._row_factory(**kw)

    def fetchmany(self, size=None):
        for row in super().fetchmany(size):
            kw = dict(zip([c.name for c in self.description], row))
            yield self._row_factory(**kw)

    def fetchall(self):
        for row in super().fetchall():
            kw = dict(zip([c.name for c in self.description], row))
            yield self._row_factory(**kw)


def scalar(**kw):
    # Row factory for scalar.
    return next(iter(kw.values()))


class ConnectionPool(ThreadedConnectionPool):
    def __init__(self, app, observers=None, **kw):
        self.app = app
        self.observers = observers
        super().__init__(**kw)

    def _connect(self, *a, **kw):
        return retry_connect(super()._connect, self.app, *a, **kw)

    def auto_reconnect(self):
        # Helper to recover a connection lost in a code block
        #
        # Manage pooled connection lost. Yield a context manager for one or two
        # attempt. The first attempt uses the connection as returned by the
        # pool. The second attempt closes pool and request a fresh connection.
        #
        # for attempt in pool.auto_reconnect():
        #     with attempt() as conn:
        #         conn.queryscalar("SELECT 1")
        #
        manager = ReconnectManager(self)
        for try_ in 0, 1:
            yield manager
            if not manager.retry:
                break

    def closeall(self):
        # Close all connection, keeping pool opened.
        for conn in self._pool + list(self._used.values()):
            conn.close()
            self.putconn(conn)


class DBConnectionPool:
    # Pool one connection per database.
    #
    # Not thread-safe.

    def __init__(self, postgres):
        self.postgres = postgres
        self.pool = dict()

    def getconn(self, dbname=None):
        dbname = dbname or self.postgres.dbname
        conn = self.pool.get(dbname)
        if not conn:
            logger.debug("Opening connection to db %s.", dbname)
            pqvars = self.postgres.pqvars(dbname=dbname)
            conn = retry_connect(connect, self.postgres.app, **pqvars)
            self.pool[dbname] = conn

        return conn

    def auto_reconnect(self, dbname=None):
        # Helper to recover a connection lost in a code block
        #
        # Manage pooled connection lost. Yield a context manager for one or two
        # attempt. The first attempt uses the connection as returned by the
        # pool. The second attempt closes pool and request a fresh connection.
        #
        # for attempt in dbpool.auto_reconnect():
        #     with attempt() as conn:
        #         conn.queryscalar("SELECT 1")
        #
        manager = ReconnectManager(self, dbname)
        for try_ in 0, 1:
            yield manager
            if not manager.retry:
                break

    def putconn(self, *_):
        # Keep a single connection open, per database. Closes only upon pool
        # closing.
        pass

    def closeall(self):
        for dbname in list(self.pool):
            conn = self.pool.pop(dbname)
            logger.debug("Closing pooled connection to %s.", dbname)
            conn.close()

    def __del__(self):
        self.closeall()

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.closeall()


def retry_connect(connect, app, *a, **kw):
    # Retry 3 times, 1 second wait between each try. This allows to wait for
    # Postgres to restart, but give up to consider Postgres as unavailable.
    for wait in [1] * 3 + [0]:
        try:
            conn = connect(*a, **kw)
            if not app.status.data["postgres"]["available"]:
                logger.info("Recovered Postgres connexion.")
                app.postgres.notify_observers()
            app.status.data["postgres"]["available"] = True
            conn.set_session(autocommit=True)
            return conn
        except Exception as e:
            app.status.data["postgres"]["available"] = False
            if wait:
                logger.debug("Retrying connection open in %ss: %s", wait, e)
                sleep(wait)
            else:  # Last wait is 0. Just give up.
                raise


class ReconnectManager:
    def __init__(self, pool, *connect_args):
        self.pool = pool
        self.connect_args = connect_args
        self.conn = None
        self.retry = False

    def __call__(self, *connect_args):
        if connect_args:
            self.connect_args = connect_args
        return self

    def __enter__(self):
        self.conn = self.pool.getconn(*self.connect_args)
        self.retry = False
        return self.conn

    def __exit__(self, exc_type, e, exc_tb):
        if isinstance(e, Psycopg2Error):
            # Class 57 is Operator Intervention like shutdown, terminate backend, etc.
            # Cf. https://www.postgresql.org/docs/current/errcodes-appendix.html#ERRCODES-TABLE
            if e.pgcode is None or e.pgcode.startswith("57"):
                logger.debug("Retrying lost connection: %s", e)
                self.pool.closeall()
                self.conn = None
                self.retry = True
                return True

        # Else, just clean conn and let exception bubble.
        self.pool.putconn(self.conn)
        self.conn = None


def extract_conninfo_fields(conninfo, searched_keys=["host", "port"]):
    """Return only searched_keys from conninfo"""
    if conninfo is None:
        return None
    if "password" in searched_keys:
        raise Exception("Password should not be fetched from conninfo.")
    filterred_conninfo = {}
    for kv in conninfo.split(" "):
        k, _, v = kv.partition("=")
        if k in searched_keys:
            filterred_conninfo[k] = v
    return filterred_conninfo



import contextlib
import functools
import inspect
import json
import logging
from datetime import timedelta

from bottle import (
    Bottle,
    HTTPError,
    HTTPResponse,
    Response,
    default_app,
    request,
    response,
)

from .. import notification
from ..toolkit.errors import TemboardError
from ..toolkit.http import format_date
from ..toolkit.signing import InvalidSignature, canonicalize_request, verify_v1
from ..toolkit.utils import JSONEncoder, utcnow

logger = logging.getLogger(__name__)


def create_app(temboard):
    Response.default_content_type = "text/plain"
    app = CustomBottle(autojson=False)
    app.temboard = temboard
    app.add_hook("before_request", before_request_log)
    # First declared, first executed.
    app.install(JSONPlugin())
    app.install(SignaturePlugin())
    app.install(PostgresPlugin())
    return app


class CustomBottle(Bottle):
    def default_error_handler(self, res):
        # Skip HTML template.

        headers = response.headers.dict.copy()
        headers["Content-Type"] = "application/json"

        # res.exception is set by the bottle's catchall
        if res.exception:
            logger.exception("Unhandled error:", exc_info=res.exception)
        if isinstance(res.body, str):
            res.body = {"error": res.body}
        return HTTPResponse(json.dumps(res.body), res.status, headers=headers)

    def mount(self, prefix, app, **options):
        for plugin in self.plugins:
            app.install(plugin)
        return super().mount(prefix, app, **options)

    def push_audit_notification(self, message):
        n = notification.Notification(request.username, message)
        # push() already logs SQLite traceback.
        with contextlib.suppress(notification.NotificationError):
            notification.NotificationMgmt.push(self.temboard.config, n)


def before_request_log():
    logger.debug("New web request: %s %s", request.method, request.path)


class PostgresPlugin:
    def __init__(self):
        self._pool = None
        self._dbpool = None

    @property
    def dbpool(self):
        if not self._dbpool:
            self._dbpool = default_app().temboard.postgres.dbpool()
        return self._dbpool

    @property
    def pool(self):
        if not self._pool:
            self._pool = default_app().temboard.postgres.pool()
        return self._pool

    def apply(self, callback, route):
        wanted = self.wants_postgres(callback)
        if not wanted:
            return callback

        @functools.wraps(callback)
        def wrapper(*a, **kw):
            if "pgpool" in wanted:
                # dbpool is not threadsafe! But wsgiref simple server is
                # not threaded.
                kw["pgpool"] = self.dbpool

            # Assume callbacks idempotence.
            for attempt in self.pool.auto_reconnect():
                with attempt() as conn:
                    if "pgconn" in wanted:
                        kw["pgconn"] = conn

                    return callback(*a, **kw)

                if attempt.retry:
                    # Propagate connection lost to dbpool.
                    self.dbpool.closeall()

        return wrapper

    def wants_postgres(self, callback):
        argspec = inspect.getfullargspec(callback)
        return [a for a in argspec.args if a in ("pgconn", "pgpool")]


class JSONPlugin:
    # Bottle jsonify only dict. JSON Array was a security issue for some
    # browser.
    def apply(self, callback, route):
        @functools.wraps(callback)
        def wrapper(*a, **kw):
            res = callback(*a, **kw)
            body = res.body if isinstance(res, HTTPResponse) else res

            if not isinstance(body, (dict, list)):
                return res

            body = json.dumps(body, cls=JSONEncoder)
            if not isinstance(res, HTTPResponse):
                res = response.copy(cls=HTTPResponse)
            res.body = body

            res.headers["Content-Type"] = "application/json"

            return res

        return wrapper


class SignaturePlugin:
    name = "signature"

    def setup(self, app):
        self.app = app

    def apply(self, callback, route):
        @functools.wraps(callback)
        def wrapper(*a, **kw):
            request.username = self.authenticate()
            return callback(*a, **kw)

        return wrapper

    def authenticate(self):
        app = default_app().temboard

        date = request.headers.get("x-temboard-date")
        if not date:
            raise HTTPError(400, "Missing X-TemBoard-Date header.")

        oldest_date = format_date(utcnow() - timedelta(hours=2))
        if date < oldest_date:
            raise HTTPError(400, "Request older than 2 hours.")

        signature = request.headers.get("x-temboard-signature")
        if not signature:
            raise HTTPError(400, "Missing X-TemBoard-Signature header.")
        version, _, signature = signature.partition(":")
        if "v1" != version:
            raise HTTPError(400, "Unsupported signature format")

        if not signature:
            raise HTTPError(400, "Malformed signature")

        path = request.environ["RAW_PATH_INFO"]
        if request.environ["QUERY_STRING"]:
            path = path + "?" + request.environ["QUERY_STRING"]
        try:
            canonical_request = canonicalize_request(
                request.method, path, request.headers, request.body.read()
            )
        except TemboardError as e:
            raise HTTPError(400, str(e))

        try:
            verify_v1(app.config.signing_key, signature, canonical_request)
        except InvalidSignature:
            raise HTTPError(403, "Invalid signature")

        user = request.headers.get("x-temboard-user")
        if not user:
            raise HTTPError(400, "Missing username")

        return user



import json
import logging
import time
from datetime import datetime

from bottle import HTTPError, default_app, request, response

from ... import __version__ as __VERSION__
from ...toolkit import taskmanager
from ...toolkit.configuration import OptionSpec
from ...toolkit.validators import commalist
from ...tools import now, validate_parameters
from ...web.app import CustomBottle
from . import db
from .inventory import host_info, instance_info
from .openmetrics import format_open_metrics_lines, generate_samples
from .output import remove_passwords
from .probes import load_probes, run_probes

logger = logging.getLogger(__name__)
bottle = CustomBottle()
workers = taskmanager.WorkerSet()

T_TIMESTAMP_UTC = b"(^[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}Z$)"
T_LIMIT = b"(^[0-9]+$)"


@bottle.get("/metrics")
def get_metrics():
    app = default_app().temboard
    response.headers["Content-Type"] = "text/plain; version=0.0.4"

    rows = db.get_metrics(app.config.temboard.home, "monitoring.db")
    if not rows:
        return "# EOF\n"
    ((_, data),) = rows
    data = json.loads(data)
    db.use_current_for_delta_metrics(data)
    lines = format_open_metrics_lines(generate_samples(data))
    return "\n".join(lines)


@bottle.get("/history")
def get_monitoring():
    """Monitoring root API aims to query metrics history.
    Data are sorted by collect timestamp, in ascending order. By default, only
    the most fresh record set is returned. The query parameter 'start' can be
    used as lower bound and must be expressed as a UTC timestamp formatted
    using ISO8601 norm with a terminal 'Z' character. To limit the number of
    returned records to N, the query parameter 'limit' can be used and set to
    N. 'limit' default value is 50, meaning that the maximum number of record
    set this API returns by default is 50.
    """

    # Default values
    start_timestamp = None
    limit = 50

    app = default_app().temboard

    if "start" in request.query:
        # Validate start parameter
        validate_parameters(request.query, [("start", T_TIMESTAMP_UTC, False)])
        # Convert it to epoch
        try:
            start_timestamp = (
                datetime.strptime(request.query["start"], "%Y-%m-%dT%H:%M:%SZ")
                - datetime(1970, 1, 1)
            ).total_seconds()
        except ValueError:
            raise HTTPError(406, "Invalid timestamp")

    if "limit" in request.query:
        # Validate limit parameter
        validate_parameters(request.query, [("limit", T_LIMIT, False)])
        limit = int(request.query["limit"])

    out = []
    h, n = app.config.temboard.home, "monitoring.db"
    for _, metrics in db.get_metrics(h, n, limit, start_timestamp):
        metrics = json.loads(metrics)
        # Dropping current value, use /metrics to get them.
        db.drop_current_for_delta_metrics(metrics)
        out.append(metrics)
    response.set_header("X-TemBoard-Discover-ETag", app.discover.etag)
    return out


@bottle.get("/config")
def get_config():
    """Returns monitoring plugin configuration."""
    app = default_app().temboard
    return dict(
        dbnames=app.config.monitoring.dbnames,
        scheduler_interval=app.config.monitoring.scheduler_interval,
    )


@workers.register(pool_size=1)
def monitoring_collector_worker(app):
    """
    Run probes and push collected metrics in a queue.
    """
    logger.info("Collecting metrics.")
    config = app.config
    logger.debug("Gathering host information.")
    discover = app.discover.ensure_latest()
    system_info = host_info(discover)
    logger.debug("Load the probes to run.")
    probes = load_probes(config.monitoring, config.temboard.home)

    with app.postgres.dbpool() as pool:
        instance = instance_info(pool, app.config.monitoring.dbnames, discover)
        data = run_probes(probes, pool, [instance])

    # Prepare and send output
    output = dict(
        datetime=now(),
        hostinfo=system_info,
        instances=remove_passwords([instance]),
        data=data,
        version=__VERSION__,
    )
    logger.debug("Add data to metrics table.")
    db.add_metric(config.temboard.home, "monitoring.db", time.time(), output)

    logger.debug("Collect done.")

    try:
        logger.debug("temboard_agent_version=%s", __VERSION__)
        logger.debug("hostinfo=%s", system_info)
        for record in iter_metrics_for_logfmt(data):
            # up=1 is a marker to grep logfmt lines
            logger.debug("up=1 %s", " ".join("%s=%s" % i for i in record.items()))
    except Exception:
        logger.exception("Failed to log metrics.")


def iter_metrics_for_logfmt(data):
    # Generates a flat sequence of record dict containing key value for logfmt
    # printing. See dev/perfui/ in temboard project to analyze such data.
    blacklist = ("current", "datetime", "port", "cpu", "measure_interval")
    for k, v in data.items():
        for vv in v:
            record = dict()
            for kkk, vvv in vv.items():
                if hasattr(vvv, "isoformat"):
                    vvv = vvv.isoformat(sep="T")
                if kkk in blacklist:
                    continue
                if kkk in ("dbname", "spcname") or k in ("loadavg", "memory"):
                    record[kkk] = vvv
                elif k.endswith("_size") and kkk == "size":
                    record[k] = vvv
                elif k == "lag":
                    record[k] = vvv
                else:
                    record[f"{k}_{kkk}"] = vvv
            yield record


class MonitoringPlugin:
    PG_MIN_VERSION = (90400, 9.4)
    s = "monitoring"
    option_specs = [
        OptionSpec(s, "dbnames", default="*", validator=commalist),
        OptionSpec(s, "scheduler_interval", default=60, validator=int),
        OptionSpec(s, "probes", default="*", validator=commalist),
    ]
    del s

    def __init__(self, app, **kw):
        self.app = app
        self.app.config.add_specs(self.option_specs)

    def load(self):
        default_app().mount("/monitoring", bottle)
        self.app.worker_pool.add(workers)
        logger.debug(
            "Schedule metric collect every %s seconds.",
            self.app.config.monitoring.scheduler_interval,
        )
        workers.schedule(
            id="monitoring_collector",
            redo_interval=self.app.config.monitoring.scheduler_interval,
        )(monitoring_collector_worker)
        self.app.scheduler.add(workers)
